"""Wandb plugin.

!!! todo

    This plugin is experimental and out of date.

"""
from __future__ import annotations

import logging
from collections.abc import Callable, Mapping
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, replace
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Concatenate,
    Generic,
    Literal,
    TypeAlias,
    TypeVar,
)
from typing_extensions import ParamSpec, Self, override

import numpy as np
import wandb
from wandb.sdk import wandb_run
from wandb.sdk.lib import RunDisabled

from amltk.scheduling.executors import SequentialExecutor
from amltk.scheduling.plugins.plugin import Plugin

if TYPE_CHECKING:
    from amltk.optimization import Trial
    from amltk.scheduling import Scheduler, Task

P = ParamSpec("P")
R = TypeVar("R")


logger = logging.getLogger(__name__)

WRun: TypeAlias = wandb_run.Run | RunDisabled


@dataclass
class WandbParams:
    """Parameters for initializing a wandb run.

    This class is a dataclass that contains all the parameters that are used
    to initialize a wandb run. It is used by the
    [`WandbPlugin`][amltk.scheduling.plugins.wandb.WandbPlugin] to initialize a run.
    It can be modified using the
    [`modify()`][amltk.scheduling.plugins.wandb.WandbParams.modify] method.

    Please refer to the documentation of the
    [`wandb.init()`](https://docs.wandb.ai/ref/python/init) method for more information
    on the parameters.
    """

    project: str | None = None
    group: str | None = None
    job_type: str | None = None
    entity: str | None = None
    tags: list[str] | None = None
    notes: str | None = None
    reinit: bool | None = None
    config_exclude_keys: list[str] | None = None
    config_include_keys: list[str] | None = None
    resume: bool | str | None = None
    mode: Literal["online", "offline", "disabled"] = "online"
    allow_val_change: bool = False
    force: bool = False
    dir: str | Path | None = None

    # TODO: There is a parameter for `id`, this should be used in multifidelity
    # runs.

    def modify(self, **kwargs: Any) -> WandbParams:
        """Modify the parameters of this instance.

        This method returns a new instance of this class with the parameters
        modified. This is useful for example when you want to modify the
        parameters of a run to add tags or notes.
        """
        return replace(self, **kwargs)

    def run(
        self,
        name: str,
        config: Mapping[str, Any] | None = None,
    ) -> WRun:
        """Initialize a wandb run.

        This method initializes a wandb run using the parameters of this
        instance. It returns the wandb run object.

        Args:
            name: The name of the run.
            config: The configuration of the run.

        Returns:
            The wandb run object.
        """
        run = wandb.init(
            config=dict(config) if config else None,
            name=name,
            project=self.project,
            group=self.group,
            tags=self.tags,
            entity=self.entity,
            notes=self.notes,
            reinit=self.reinit,
            dir=self.dir,
            config_exclude_keys=self.config_exclude_keys,
            config_include_keys=self.config_include_keys,
            mode=self.mode,
            allow_val_change=self.allow_val_change,
            force=self.force,
        )
        if run is None:
            raise RuntimeError("Wandb run was not initialized")

        return run


class WandbLiveRunWrap(Generic[P]):
    """Wrap a function to log the results to a wandb run.

    This class is used to wrap a function that returns a report to log the
    results to a wandb run. It is used by the
    [`WandbTrialTracker`][amltk.scheduling.plugins.wandb.WandbTrialTracker] to wrap
    the target function.
    """

    def __init__(
        self,
        params: WandbParams,
        fn: Callable[Concatenate[Trial, P], Trial.Report],
        *,
        modify: Callable[[Trial, WandbParams], WandbParams] | None = None,
    ):
        """Initialize the wrapper.

        Args:
            params: The parameters to initialize the wandb run.
            fn: The function to wrap.
            modify: A function that modifies the parameters of the wandb run
                before each trial.
        """
        super().__init__()
        self.params = params
        self.fn = fn
        self.modify = modify

    def __call__(self, trial: Trial, *args: P.args, **kwargs: P.kwargs) -> Trial.Report:
        """Call the wrapped function and log the results to a wandb run."""
        params = self.params if self.modify is None else self.modify(trial, self.params)
        with params.run(name=trial.name, config=trial.config) as run:
            # Make sure the run is available from the trial
            trial.extras["wandb"] = run

            report = self.fn(trial, *args, **kwargs)

            report_df = report.df()
            run.log({"table": wandb.Table(dataframe=report_df)})
            wandb_summary = {
                k: v
                for k, v in report.summary.items()
                if isinstance(v, int | float | np.number)
            }
            run.summary.update(wandb_summary)

        wandb.finish()
        return report


class WandbTrialTracker(Plugin):
    """Track trials using wandb.

    This class is a task plugin that tracks trials using wandb.
    """

    name: ClassVar[str] = "wandb-trial-tracker"
    """The name of the plugin."""

    def __init__(
        self,
        params: WandbParams,
        *,
        modify: Callable[[Trial, WandbParams], WandbParams] | None = None,
    ):
        """Initialize the plugin.

        Args:
            params: The parameters to initialize the wandb run.
            modify: A function that modifies the parameters of the wandb run
                before each trial.
        """
        super().__init__()
        self.params = params
        self.modify = modify

    @override
    def attach_task(self, task: Task) -> None:
        """Use the task to register several callbacks."""
        self._check_explicit_reinit_arg_with_executor(task.scheduler)

    @override
    def pre_submit(
        self,
        fn: Callable[P, R],
        *args: Any,
        **kwargs: Any,
    ) -> tuple[Callable[P, R], tuple, dict] | None:
        """Wrap the target function to log the results to a wandb run.

        This method wraps the target function to log the results to a wandb run
        and returns the wrapped function.

        Args:
            fn: The target function.
            args: The positional arguments of the target function.
            kwargs: The keyword arguments of the target function.

        Returns:
            The wrapped function, the positional arguments and the keyword
            arguments.
        """
        fn = WandbLiveRunWrap(self.params, fn, modify=self.modify)  # type: ignore
        return fn, args, kwargs

    @override
    def copy(self) -> Self:
        """Copy the plugin."""
        return self.__class__(modify=self.modify, params=replace(self.params))

    def _check_explicit_reinit_arg_with_executor(
        self,
        scheduler: Scheduler,
    ) -> None:
        """Check if reinit arg was explicitly set and conflicts with executor.

        Args:
            scheduler: The scheduler to check.
        """
        if isinstance(scheduler.executor, SequentialExecutor | ThreadPoolExecutor):
            if self.params.reinit is False:
                raise ValueError(
                    "WandbPlugin reinit argument is not compatible with"
                    " SequentialExecutor or ThreadPoolExecutor. This is because"
                    " wandb.init(reinit=False) will only work if each trial is"
                    " run in a separate process. If you want to use `reinit=False`,"
                    " please consider using a different executor. Otherwise, you can"
                    " also set `reinit=True` to explicitly allow reinitialization."
                    " By default `reinit` is `None`, which means that it will be"
                    " set to `True` if the executor is not a SequentialExecutor or"
                    " ThreadPoolExecutor, and `False` otherwise.",
                )

            self.params.reinit = True


class WandbPlugin:
    """Log trials using wandb.

    This class is the entry point to log trials using wandb. It
    can be used to create a
    [`trial_tracker()`][amltk.scheduling.plugins.wandb.WandbPlugin.trial_tracker]
    to pass into a [`Task(plugins=...)`][amltk.Task] or to
    create `wandb.Run`'s for custom purposes with
    [`run()`][amltk.scheduling.plugins.wandb.WandbPlugin.run].
    """

    def __init__(
        self,
        *,
        project: str,
        group: str | None = None,
        entity: str | None = None,
        dir: str | Path | None = None,  # noqa: A002
        mode: Literal["online", "offline", "disabled"] = "online",
    ):
        """Initialize the plugin.

        Args:
            project: The name of the project.
            group: The name of the group.
            entity: The name of the entity.
            dir: The directory to store the runs in.
            mode: The mode to use for the runs.
        """
        super().__init__()
        _dir = Path(project) if dir is None else Path(dir)
        _dir.mkdir(parents=True, exist_ok=True)

        self.dir = _dir.resolve().absolute()
        self.project = project
        self.group = group
        self.entity = entity
        self.mode = mode

    def trial_tracker(
        self,
        job_type: str = "trial",
        *,
        modify: Callable[[Trial, WandbParams], WandbParams] | None = None,
    ) -> WandbTrialTracker:
        """Create a live tracker.

        Args:
            job_type: The job type to use for the runs.
            modify: A function that modifies the parameters of the wandb run
                before each trial.

        Returns:
            A live tracker.
        """
        params = WandbParams(
            project=self.project,
            entity=self.entity,
            group=self.group,
            dir=self.dir,
            mode=self.mode,  # type: ignore
            job_type=job_type,
        )
        return WandbTrialTracker(params, modify=modify)

    def run(
        self,
        *,
        name: str,
        job_type: str | None = None,
        group: str | None = None,
        config: Mapping[str, Any] | None = None,
        tags: list[str] | None = None,
        resume: bool | str | None = None,
        notes: str | None = None,
    ) -> WRun:
        """Create a wandb run.

        See [`wandb.init()`](https://docs.wandb.ai/ref/python/init) for more.
        """
        return WandbParams(
            project=self.project,
            entity=self.entity,
            group=group,
            dir=self.dir,
            mode=self.mode,  # type: ignore
            job_type=job_type,
            tags=tags,
            resume=resume,
            notes=notes,
        ).run(
            name=name,
            config=config,
        )



"""Custom estimators for use with scikit-learn."""

# TODO:
#   * Document Stored Prediction Estimators
from __future__ import annotations

from typing import TYPE_CHECKING, Any
from typing_extensions import Self

from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin

from amltk.data.conversions import probabilities_to_classes

if TYPE_CHECKING:
    import numpy as np


class StoredPredictionRegressor(BaseEstimator, RegressorMixin):
    """A class that just uses precomputed values for regression."""

    def __init__(self, predictions: np.ndarray):
        """Initialize the estimator.

        Args:
            predictions: The precomputed predictions.
        """
        super().__init__()
        self.predictions = predictions

    def fit(self, *_: Any, **__: Any) -> Self:
        """Fit the estimator. Doesn't do anything."""
        return self

    def predict(self, X: Any, *_: Any, **__: Any) -> np.ndarray:  # noqa: N803, ARG002
        """Predict the target values, returning the precomputed values."""
        return self.predictions


class StoredPredictionClassifier(BaseEstimator, ClassifierMixin):
    """A class that just uses precomputed values for classification."""

    def __init__(
        self,
        predictions: np.ndarray | None = None,
        probabilities: np.ndarray | None = None,
        classes: list[np.ndarray] | np.ndarray | None = None,
    ):
        """Initialize the estimator.

        Args:
            predictions: The precomputed predictions, if any.
            probabilities: The precomputed probabilities, if any.
            classes: The classes, if any.
        """
        super().__init__()
        self.predictions = predictions
        self.probabilities = probabilities
        self.classes = classes

        # HACK: This is to enable sklearn-compatibility
        # `clone` and other methods rely on this trailing underscore
        # to indicate fitted attributes. We essentially declare it fitted
        # at init for simplicity
        self.classes_ = classes

    def fit(self, *_: Any, **__: Any) -> Self:
        """Fit the estimator. Doesn't do anything."""
        return self

    def predict(self, X: Any, *_: Any, **__: Any) -> np.ndarray:  # noqa: N803, ARG002
        """Predict the target values, returning the precomputed values."""
        if self.predictions is None:
            if self.probabilities is None:
                raise RuntimeError(
                    "No predictions or probabilities were provided during",
                    " construction, so this estimator cannot be used for",
                    " `predict()`.",
                )
            if self.classes_ is None:
                raise RuntimeError(
                    "No classes were provided during construction, so it can't"
                    " be used for `predict()` from probabilities.",
                )

            predictions = probabilities_to_classes(
                self.probabilities,
                classes=self.classes_,
            )
        else:
            predictions = self.predictions

        return predictions

    def predict_proba(
        self,
        X: Any,  # noqa: N803, ARG002
        *_: Any,
        **__: Any,
    ) -> np.ndarray:
        """Predict the probabilities, returning the precomputed values."""
        if self.probabilities is None:
            raise RuntimeError(
                "No probabilities were provided during construction, so this"
                " estimator cannot be used for `predict_proba()`.",
            )

        return self.probabilities



"""The sklearn [`builder()`][amltk.pipeline.builders.sklearn.build], converts
a pipeline made of [`Node`][amltk.pipeline.Node]s into a sklearn
[`Pipeline`][sklearn.pipeline.Pipeline].

!!! tip "Requirements"

    This requires `sklearn` which can be installed with:

    ```bash
    pip install "amltk[scikit-learn]"

    # Or directly
    pip install scikit-learn
    ```


??? tip "Basic Usage"

    ```python
    # TODO
    ```


Each _kind_ of node corresponds to a different part of the end pipeline:

=== "`Fixed`"

    [`Fixed`][amltk.pipeline.Fixed] - The estimator will simply be cloned, allowing you
    to directly configure some object in a pipeline.

    ```python exec="true" source="material-block" html="true"
    from sklearn.ensemble import RandomForestClassifier
    from amltk.pipeline import Fixed

    est = Fixed(RandomForestClassifier(n_estimators=25))
    built_pipeline = est.build("sklearn")
    from amltk._doc import doc_print; doc_print(print, built_pipeline)  # markdown-exec: hide
    ```

=== "`Component`"

    [`Component`][amltk.pipeline.Component] - The estimator will be built from the
    component's config. This is mostly useful to allow a space to be defined for
    the component.

    ```python exec="true" source="material-block" html="true"
    from sklearn.ensemble import RandomForestClassifier
    from amltk.pipeline import Component

    est = Component(RandomForestClassifier, space={"n_estimators": (10, 100)})

    # ... Likely get the configuration through an optimizer or sampling
    configured_est = est.configure({"n_estimators": 25})

    built_pipeline = configured_est.build("sklearn")
    from amltk._doc import doc_print; doc_print(print, built_pipeline)  # markdown-exec: hide
    ```

=== "`Sequential`"

    [`Sequential`][amltk.pipeline.Sequential] - The sequential will be converted into a
    [`Pipeline`][sklearn.pipeline.Pipeline], building whatever nodes are contained
    within in.

    ```python exec="true" source="material-block" html="true"
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.decomposition import PCA
    from amltk.pipeline import Component, Sequential

    pipeline = Sequential(
        PCA(n_components=3),
        Component(RandomForestClassifier, config={"n_estimators": 25})
    )
    built_pipeline = pipeline.build("sklearn")
    from amltk._doc import doc_print; doc_print(print, built_pipeline)  # markdown-exec: hide
    ```

=== "`Split`"

    [`Split`][amltk.pipeline.Split] - The split will be converted into a
    [`ColumnTransformer`][sklearn.compose.ColumnTransformer], where each path
    and the data that should go through it is specified by the split's config.
    You can provide a `ColumnTransformer` directly as the item to the `Split`,
    or otherwise if left blank, it will default to the standard sklearn one.

    You can use a `Fixed` with the special keyword `"passthrough"` as you might normally
    do with a `ColumnTransformer`.

    By default, we provide two special keywords you can provide to a `Split`,
    namely `#!python "categorical"` and `#!python "numerical"`, which will
    automatically configure a `ColumnTransorfmer` to pass the appropraite
    columns of a data-frame to the given paths.

    ```python exec="true" source="material-block" html="true" hl_lines="16"
    from amltk.pipeline import Split, Component
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.impute import SimpleImputer
    from sklearn.preprocessing import StandardScaler, OneHotEncoder

    categorical_pipeline = [
        SimpleImputer(strategy="constant", fill_value="missing"),
        Component(
            OneHotEncoder,
            space={
                "min_frequency": (0.01, 0.1),
                "handle_unknown": ["ignore", "infrequent_if_exist"],
            },
            config={"drop": "first"},
        ),
    ]
    numerical_pipeline = [SimpleImputer(strategy="median"), StandardScaler()]

    split = Split(
        {
            "categories": categorical_pipeline,
            "numbers": numerical_pipeline
        }
    )
    from amltk._doc import doc_print; doc_print(print, split)  # markdown-exec: hide
    ```

    You can manually specify the column selectors if you prefer.

    ```python
    split = Split(
        {
            "categories": categorical_pipeline,
            "numbers": numerical_pipeline,
        },
        config={
            "categories": make_column_selector(dtype_include=object),
            "numbers": make_column_selector(dtype_include=np.number),
        },
   )
    ```



=== "`Join`"

    [`Join`][amltk.pipeline.Join] - The join will be converted into a
    [`FeatureUnion`][sklearn.pipeline.FeatureUnion].

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Join, Component
    from sklearn.decomposition import PCA
    from sklearn.feature_selection import SelectKBest

    join = Join(PCA(n_components=2), SelectKBest(k=3), name="my_feature_union")

    pipeline = join.build("sklearn")
    from amltk._doc import doc_print; doc_print(print, pipeline)  # markdown-exec: hide
    ```

=== "`Choice`"

    [`Choice`][amltk.pipeline.Choice] - The estimator will be built from the chosen
    component's config. This is very similar to [`Component`][amltk.pipeline.Component].

    ```python exec="true" source="material-block" html="true"
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.neural_network import MLPClassifier
    from amltk.pipeline import Choice

    # The choice here is usually provided during the `.configure()` step.
    estimator_choice = Choice(
        RandomForestClassifier(),
        MLPClassifier(),
        config={"__choice__": "RandomForestClassifier"}
    )

    built_pipeline = estimator_choice.build("sklearn")
    from amltk._doc import doc_print; doc_print(print, built_pipeline)  # markdown-exec: hide
    ```


"""  # noqa: E501
from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, TypeVar

from sklearn.base import BaseEstimator, clone
from sklearn.compose import ColumnTransformer, make_column_selector
from sklearn.pipeline import (
    FeatureUnion,
    Pipeline as SklearnPipeline,
)

from amltk.pipeline import (
    Choice,
    Component,
    Fixed,
    Join,
    Node,
    Searchable,
    Sequential,
    Split,
)

if TYPE_CHECKING:
    from typing import TypeAlias

COLUMN_TRANSFORMER_ARGS = [
    "remainder",
    "sparse_threshold",
    "n_jobs",
    "transformer_weights",
    "verbose",
    "verbose_feature_names_out",
]
FEATURE_UNION_ARGS = ["n_jobs", "transformer_weights", "verbose"]

# TODO: We can make this more explicit with typing out sklearn types.
# However sklearn operates in a bit more of a general level so it would
# require creating protocols to type this properly and work with sklearn's
# duck-typing.
SklearnItem: TypeAlias = Any | ColumnTransformer
SklearnPipelineT = TypeVar("SklearnPipelineT", bound=SklearnPipeline)


def _process_split(
    node: Split,
    pipeline_type: type[SklearnPipelineT] = SklearnPipeline,
    **pipeline_kwargs: Any,
) -> tuple[str, ColumnTransformer]:
    if any(child.name in COLUMN_TRANSFORMER_ARGS for child in node.nodes):
        raise ValueError(
            f"Can't handle step as it has a path with a name that matches"
            f" a known ColumnTransformer argument: {node}",
        )

    config = dict(node.config) if node.config is not None else {}

    # Automatic categories/numerical config
    for keyword, dtypes in (
        ("categorical", [object, "category"]),
        ("numerical", ["number"]),
    ):
        if any(child.name == keyword for child in node.nodes) and keyword not in config:
            config[keyword] = make_column_selector(dtype_include=dtypes)

    # Automatic numeric config
    if any(child.name not in config for child in node.nodes):
        raise ValueError(
            f"Can't handle split {node.name=} as some path has no config associated"
            " with it."
            "\nPlease ensure that all paths have a config associated with them."
            f"\n{config=}\n"
            f"children={[child.name for child in node.nodes]}\n",
        )

    match node.item:
        case None:
            col_transform_cls = ColumnTransformer
        case type() if issubclass(node.item, ColumnTransformer):
            col_transform_cls = node.item
        case _:
            raise ValueError(
                f"Can't handle: {node}.\n"
                " Requires all splits to have a subclass"
                " ColumnTransformer as the item, or None.",
            )

    # https://scikit-learn.org/stable/modules/generated/sklearn.compose.ColumnTransformer.html
    # list of (name, estimator, columns)
    transformers: list[tuple[str, Any, Any]] = []

    for child in node.nodes:
        child_steps = list(_iter_steps(child))
        match child_steps:
            case []:
                raise ValueError(f"Can't handle child of split.\n{child=}\n{node}")
            case [(name, sklearn_thing)]:
                transformers.append((name, sklearn_thing, config[child.name]))
            case list():
                sklearn_thing = pipeline_type(child_steps, **pipeline_kwargs)
                transformers.append((child.name, sklearn_thing, config[child.name]))

    return (node.name, col_transform_cls(transformers))


def _process_join(
    node: Join,
    pipeline_type: type[SklearnPipelineT] = SklearnPipeline,
    **pipeline_kwargs: Any,
) -> tuple[str, FeatureUnion]:
    if any(child.name in FEATURE_UNION_ARGS for child in node.nodes):
        raise ValueError(
            f"Can't handle step as it has a path with a name that matches"
            f" a known FeatureUnion argument: {node}",
        )

    match node.item:
        case None:
            feature_union_cls = FeatureUnion
        case type() if issubclass(node.item, FeatureUnion):
            feature_union_cls = node.item
        case _:
            raise ValueError(
                f"Can't handle: {node}.\n"
                " Requires all splits to have a subclass"
                " ColumnTransformer as the item, or None.",
            )

    # https://scikit-learn.org/stable/modules/generated/sklearn.pipeline.FeatureUnion.html
    # list of (name, estimator)
    transformers: list[tuple[str, Any]] = []

    for child in node.nodes:
        child_steps = list(_iter_steps(child))
        match child_steps:
            case []:
                raise ValueError(f"Can't handle child of Join.\n{child=}\n{node}")
            case [(name, sklearn_thing)]:
                transformers.append((name, sklearn_thing))
            case list():
                sklearn_thing = pipeline_type(child_steps, **pipeline_kwargs)
                transformers.append((child.name, sklearn_thing))

    return (node.name, feature_union_cls(transformers))


def _iter_steps(
    node: Node,
    pipeline_type: type[SklearnPipelineT] = SklearnPipeline,
    **pipeline_kwargs: Any,
) -> Iterator[tuple[str, SklearnItem]]:
    match node:
        case Fixed(item=BaseEstimator()):
            yield (node.name, clone(node.item))
        case Fixed(item=anything):
            yield (node.name, anything)
        case Component():
            yield (node.name, node.build_item())
        case Choice():
            yield from _iter_steps(node.chosen())
        case Sequential():
            for child in node.nodes:
                yield from _iter_steps(child)
        # Bit more involved, we defer to another functino
        case Join():
            yield _process_join(node, pipeline_type=pipeline_type, **pipeline_kwargs)
        case Split():
            yield _process_split(node, pipeline_type=pipeline_type, **pipeline_kwargs)
        case Searchable():
            raise ValueError(f"Can't handle Searchable: {node}")
        case _:
            raise ValueError(f"Can't handle node: {node}")


def build(
    node: Node[Any, Any],
    *,
    pipeline_type: type[SklearnPipelineT] = SklearnPipeline,
    **pipeline_kwargs: Any,
) -> SklearnPipelineT:
    """Build a pipeline into a usable object.

    Args:
        node: The node from which to build a pipeline.
        pipeline_type: The type of pipeline to build. Defaults to the standard
            sklearn pipeline but can be any derivative of that, i.e. ImbLearn's
            pipeline.
        **pipeline_kwargs: The kwargs to pass to the pipeline_type.

    Returns:
        The built pipeline
    """
    return pipeline_type(list(_iter_steps(node)), **pipeline_kwargs)  # type: ignore



from __future__ import annotations

import runpy
from pathlib import Path

import pytest

HERE = Path(__file__).parent.resolve()
EXAMPLE_DIR = HERE.parent / "examples"
example_files = EXAMPLE_DIR.rglob("*.py")

parameters = [
    pytest.param(example, id=str(example.relative_to(EXAMPLE_DIR)))
    for example in example_files
]


@pytest.mark.parametrize("example", parameters)
@pytest.mark.example()
def test_example(example: Path) -> None:
    runpy.run_path(str(example), run_name="__main__")



"""Module for timing things."""
from __future__ import annotations

import time
from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, Literal
from typing_extensions import override

import numpy as np
import pandas as pd

from amltk._functional import dict_get_not_none

if TYPE_CHECKING:
    from pandas._libs.missing import NAType


@dataclass
class Timer:
    """A timer for measuring the time between two events.

    Attributes:
        start_time: The time at which the timer was started.
        kind: The method of timing.
    """

    start_time: float
    kind: Timer.Kind | NAType

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]) -> Timer.Interval:
        """Create a time interval from a dictionary."""
        return Timer.Interval(
            start=dict_get_not_none(d, "start", np.nan),
            end=dict_get_not_none(d, "end", np.nan),
            kind=Timer.Kind.from_str(dict_get_not_none(d, "kind", pd.NA)),
            unit=Timer.Unit.from_str(dict_get_not_none(d, "unit", pd.NA)),
        )

    @classmethod
    def na(cls) -> Timer.Interval:
        """Create a time interval with all values set to `None`."""
        return Timer.Interval(
            start=np.nan,
            end=np.nan,
            kind=pd.NA,
            unit=pd.NA,
        )

    @classmethod
    @contextmanager
    def time(
        cls,
        kind: Timer.Kind | Literal["cpu", "wall", "process"] = "wall",
    ) -> Iterator[Interval]:
        """Time a block of code.

        Args:
            kind: The type of timer to use.

        Yields:
            The Time Interval. The start and end times will not be
            valid until the context manager is exited.
        """
        timer = cls.start(kind=kind)

        interval = Timer.na()
        interval.kind = timer.kind
        interval.unit = Timer.Unit.SECONDS

        try:
            yield interval
        finally:
            _interval = timer.stop()
            interval.start = _interval.start
            interval.end = _interval.end

    @classmethod
    def start(
        cls,
        kind: Timer.Kind | Literal["cpu", "wall", "process"] = "wall",
    ) -> Timer:
        """Start a timer.

        Args:
            kind: The type of timer to use.

        Returns:
            The timer.
        """
        if kind in (Timer.Kind.WALL, "wall"):
            return Timer(time.time(), Timer.Kind.WALL)

        if kind in (Timer.Kind.CPU, "cpu"):
            return Timer(time.perf_counter(), Timer.Kind.CPU)

        if kind in (Timer.Kind.PROCESS, "process"):
            return Timer(time.process_time(), Timer.Kind.PROCESS)

        raise ValueError(f"Unknown timer type: {kind}")

    def stop(self) -> Interval:
        """Stop the timer.

        Returns:
            A tuple of the start time, end time, and duration.
        """
        if self.kind == Timer.Kind.WALL:
            end = time.time()
            return Timer.Interval(
                self.start_time,
                end,
                Timer.Kind.WALL,
                Timer.Unit.SECONDS,
            )

        if self.kind == Timer.Kind.CPU:
            end = time.perf_counter()
            return Timer.Interval(
                self.start_time,
                end,
                Timer.Kind.CPU,
                Timer.Unit.SECONDS,
            )

        if self.kind == Timer.Kind.PROCESS:
            end = time.process_time()
            return Timer.Interval(
                self.start_time,
                end,
                Timer.Kind.PROCESS,
                Timer.Unit.SECONDS,
            )

        raise ValueError(f"Unknown timer type: {self.kind}")

    @dataclass
    class Interval:
        """A time interval.

        Attributes:
            start: The start time.
            end: The end time.
            kind: The type of timer used.
            unit: The unit of time.
        """

        start: float
        end: float
        kind: Timer.Kind | NAType
        unit: Timer.Unit | NAType

        @property
        def duration(self) -> float:
            """The duration of the time interval."""
            return self.end - self.start

        def to_dict(self, *, prefix: str = "") -> dict[str, Any]:
            """Convert the time interval to a dictionary."""
            return {
                f"{prefix}start": self.start,
                f"{prefix}end": self.end,
                f"{prefix}duration": self.duration,
                f"{prefix}kind": self.kind,
                f"{prefix}unit": self.unit,
            }

    class Kind(str, Enum):
        """An enum for the type of timer."""

        WALL = "wall"
        CPU = "cpu"
        PROCESS = "process"

        @override
        def __str__(self) -> str:
            return self.name.lower()

        @override
        def __repr__(self) -> str:
            return self.name.lower()

        @classmethod
        def from_str(cls, key: Any) -> Timer.Kind | NAType:
            """Get the enum value from a string.

            Args:
                key: The string to convert.

            Returns:
                The enum value.
            """
            if isinstance(key, Timer.Kind):
                return key

            if isinstance(key, str):
                try:
                    return Timer.Kind[key.upper()]
                except KeyError:
                    return pd.NA

            return pd.NA

    class Unit(str, Enum):
        """An enum for the units of time."""

        SECONDS = "seconds"
        MILLISECONDS = "milliseconds"
        MICROSECONDS = "microseconds"
        NANOSECONDS = "nanoseconds"

        @override
        def __str__(self) -> str:
            return self.name.lower()

        @override
        def __repr__(self) -> str:
            return self.name.lower()

        @classmethod
        def from_str(cls, key: Any) -> Timer.Unit | NAType:
            """Get the enum value from a string.

            Args:
                key: The string to convert.

            Returns:
                The enum value.
            """
            if isinstance(key, Timer.Unit):
                return key

            if isinstance(key, str):
                try:
                    return Timer.Unit[key.upper()]
                except KeyError:
                    return pd.NA

            return pd.NA



from amltk.optimization.history import History
from amltk.optimization.metric import Metric
from amltk.optimization.optimizer import Optimizer
from amltk.optimization.trial import Trial

__all__ = [
    "Optimizer",
    "Trial",
    "Metric",
    "History",
]



"""The [`Scheduler`][amltk.scheduling.Scheduler] uses
an [`Executor`][concurrent.futures.Executor], a builtin python native with
a `#!python submit(f, *args, **kwargs)` function to submit compute to
be compute else where, whether it be locally or remotely.

The `Scheduler` is primarily used to dispatch compute to an `Executor` and
emit `@events`, which can trigger user callbacks.

Typically you should not use the `Scheduler` directly for dispatching and
responding to computed functions, but rather use a [`Task`][amltk.scheduling.Task]

??? note "Running in a Jupyter Notebook/Colab"

    If you are using a Jupyter Notebook, you likley need to use the following
    at the top of your notebook:

    ```python
    import nest_asyncio  # Only necessary in Notebooks
    nest_asyncio.apply()

    scheduler.run(...)
    ```

    This is due to the fact a notebook runs in an async context. If you do not
    wish to use the above snippet, you can instead use:

    ```python
    await scheduler.async_run(...)
    ```

??? tip "Basic Usage"

    In this example, we create a scheduler that uses local processes as
    workers. We then create a task that will run a function `fn` and submit it
    to the scheduler. Lastly, a callback is registered to `@on_future_result` to print the
    result when the compute is done.

    ```python exec="true" source="material-block" html="true"
    from amltk.scheduling import Scheduler

    def fn(x: int) -> int:
        return x + 1
    from amltk._doc import make_picklable; make_picklable(fn)  # markdown-exec: hide

    scheduler = Scheduler.with_processes(1)

    @scheduler.on_start
    def launch_the_compute():
        scheduler.submit(fn, 1)

    @scheduler.on_future_result
    def callback(future, result):
        print(f"Result: {result}")

    scheduler.run()
    from amltk._doc import doc_print; doc_print(print, scheduler)  # markdown-exec: hide
    ```

    The last line in the previous example called
    [`scheduler.run()`][amltk.scheduling.Scheduler.run] is what starts the scheduler
    running, in which it will first emit the `@on_start` event. This triggered the
    callback `launch_the_compute()` which submitted the function `fn` with the
    arguments `#!python 1`.

    The scheduler then ran the compute and waited for it to complete, emitting the
    `@on_future_result` event when it was done successfully. This triggered the callback
    `callback()` which printed the result.

    At this point, there is no more compute happening and no more events to respond to
    so the scheduler will halt.

??? example "`@events`"

    === "Scheduler Status Events"

        When the scheduler enters some important state, it will emit an event
        to let you know.

        === "`@on_start`"

            ::: amltk.scheduling.Scheduler.on_start

        === "`@on_finishing`"

            ::: amltk.scheduling.Scheduler.on_finishing

        === "`@on_finished`"

            ::: amltk.scheduling.Scheduler.on_finished

        === "`@on_stop`"

            ::: amltk.scheduling.Scheduler.on_stop

        === "`@on_timeout`"

            ::: amltk.scheduling.Scheduler.on_timeout

        === "`@on_empty`"

            ::: amltk.scheduling.Scheduler.on_empty

    === "Submitted Compute Events"

        When any compute goes through the `Scheduler`, it will emit an event
        to let you know. You should however prefer to use a
        [`Task`][amltk.scheduling.Task] as it will emit specific events
        for the task at hand, and not all compute.

        === "`@on_future_submitted`"

            ::: amltk.scheduling.Scheduler.on_future_submitted

        === "`@on_future_result`"

            ::: amltk.scheduling.Scheduler.on_future_result

        === "`@on_future_exception`"

            ::: amltk.scheduling.Scheduler.on_future_exception

        === "`@on_future_done`"

            ::: amltk.scheduling.Scheduler.on_future_done

        === "`@on_future_cancelled`"

            ::: amltk.scheduling.Scheduler.on_future_cancelled


??? tip "Common usages of `run()`"

    There are various ways to [`run()`][amltk.scheduling.Scheduler.run] the
    scheduler, notably how long it should run with `timeout=` and also how
    it should react to any exception that may have occurred within the `Scheduler`
    itself or your callbacks.

    Please see the [`run()`][amltk.scheduling.Scheduler.run] API doc for more
    details and features, however we show two common use cases of using the `timeout=`
    parameter.

    You can render a live display using [`run(display=...)`][amltk.scheduling.Scheduler.run].
    This require [`rich`](https://github.com/Textualize/rich) to be installed. You
    can install this with `#!bash pip install rich` or `#!bash pip install amltk[rich]`.


    === "`run(timeout=...)`"

        You can tell the `Scheduler` to stop after a certain amount of time
        with the `timeout=` argument to [`run()`][amltk.scheduling.Scheduler.run].

        This will also trigger the `@on_timeout` event as seen in the `Scheduler` output.

        ```python exec="true" source="material-block" html="True" hl_lines="19"
        import time
        from amltk.scheduling import Scheduler

        scheduler = Scheduler.with_processes(1)

        def expensive_function() -> int:
            time.sleep(0.1)
            return 42
        from amltk._doc import make_picklable; make_picklable(expensive_function)  # markdown-exec: hide

        @scheduler.on_start
        def submit_calculations() -> None:
            scheduler.submit(expensive_function)

        # The will endlessly loop the scheduler
        @scheduler.on_future_done
        def submit_again(future: Future) -> None:
            if scheduler.running():
                scheduler.submit(expensive_function)

        scheduler.run(timeout=1)  # End after 1 second
        from amltk._doc import doc_print; doc_print(print, scheduler, output="html", fontsize="small")  # markdown-exec: hide
        ```

    === "`run(timeout=..., wait=False)`"

        By specifying that the `Scheduler` should not wait for ongoing tasks
        to finish, the `Scheduler` will attempt to cancel and possibly terminate
        any running tasks.

        ```python exec="true" source="material-block" html="True"
        import time
        from amltk.scheduling import Scheduler

        scheduler = Scheduler.with_processes(1)

        def expensive_function() -> None:
            time.sleep(10)

        from amltk._doc import make_picklable; make_picklable(expensive_function)  # markdown-exec: hide

        @scheduler.on_start
        def submit_calculations() -> None:
            scheduler.submit(expensive_function)

        scheduler.run(timeout=1, wait=False)  # End after 1 second
        from amltk._doc import doc_print; doc_print(print, scheduler, output="html", fontsize="small")  # markdown-exec: hide
        ```

        ??? info "Forcibly Terminating Workers"

            As an `Executor` does not provide an interface to forcibly
            terminate workers, we provide `Scheduler(terminate=...)` as a custom
            strategy for cleaning up a provided executor. It is not possible
            to terminate running thread based workers, for example using
            `ThreadPoolExecutor` and any Executor using threads to spawn
            tasks will have to wait until all running tasks are finish
            before python can close.

            It's likely `terminate` will trigger the `EXCEPTION` event for
            any tasks that are running during the shutdown, **not***
            a cancelled event. This is because we use a
            [`Future`][concurrent.futures.Future]
            under the hood and these can not be cancelled once running.
            However there is no guarantee of this and is up to how the
            `Executor` handles this.

??? example "Scheduling something to be run later"

    You can schedule some function to be run later using the
    [`#!python scheduler.call_later()`][amltk.scheduling.Scheduler.call_later] method.

    !!! note

        This does not run the function in the background, it just schedules some
        function to be called later, where you could perhaps then use submit to
        scheduler a [`Task`][amltk.scheduling.Task] to run the function in the
        background.

    ```python exec="true" source="material-block" result="python"
    from amltk.scheduling import Scheduler

    scheduler = Scheduler.with_processes(1)

    def fn() -> int:
        print("Ending now!")
        scheduler.stop()

    @scheduler.on_start
    def schedule_fn() -> None:
        scheduler.call_later(1, fn)

    scheduler.run(end_on_empty=False)
    ```

"""  # noqa: E501
from __future__ import annotations

import asyncio
import logging
import warnings
from asyncio import Future
from collections.abc import Callable, Iterable
from concurrent.futures import Executor, ProcessPoolExecutor
from dataclasses import dataclass
from enum import Enum, auto
from functools import partial
from threading import Timer
from typing import (
    TYPE_CHECKING,
    Any,
    Concatenate,
    Literal,
    ParamSpec,
    TypeVar,
    overload,
)
from typing_extensions import Self
from uuid import uuid4

from amltk._asyncm import ContextEvent
from amltk._functional import Flag
from amltk._richutil.renderable import RichRenderable
from amltk.exceptions import SchedulerNotRunningError
from amltk.scheduling.events import Emitter, Event, Subscriber
from amltk.scheduling.executors import SequentialExecutor
from amltk.scheduling.task import Task
from amltk.scheduling.termination_strategies import termination_strategy

if TYPE_CHECKING:
    from multiprocessing.context import BaseContext

    from rich.console import RenderableType
    from rich.live import Live

    from amltk.scheduling.executors.dask_jobqueue import DJQ_NAMES
    from amltk.scheduling.plugins import Plugin
    from amltk.scheduling.plugins.comm import Comm

    P = ParamSpec("P")
    R = TypeVar("R")


logger = logging.getLogger(__name__)


class Scheduler(RichRenderable):
    """A scheduler for submitting tasks to an Executor."""

    executor: Executor
    """The executor to use to run tasks."""

    emitter: Emitter
    """The emitter to use for events."""

    queue: dict[Future, tuple[Callable, tuple, dict]]
    """The queue of tasks running."""

    on_start: Subscriber[[]]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when the
    scheduler starts. This is the first event emitted by the scheduler and
    one of the only ways to submit the initial compute to the scheduler.

    ```python
    @scheduler.on_start
    def my_callback():
        ...
    ```
    """
    on_future_submitted: Subscriber[Future]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when
    some compute is submitted.

    ```python
    @scheduler.on_future_submitted
    def my_callback(future: Future):
        ...
    ```
    """
    on_future_done: Subscriber[Future]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when
    some compute is done, regardless of whether it was successful or not.

    ```python
    @scheduler.on_future_done
    def my_callback(future: Future):
        ...
    ```
    """
    on_future_result: Subscriber[Future, Any]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when
    a future returned with a result, no exception raise.

    ```python
    @scheduler.on_future_result
    def my_callback(future: Future, result: Any):
        ...
    ```
    """
    on_future_exception: Subscriber[Future, BaseException]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when
    some compute raised an uncaught exception.

    ```python
    @scheduler.on_future_exception
    def my_callback(future: Future, exception: BaseException):
        ...
    ```
    """
    on_future_cancelled: Subscriber[Future]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called
    when a future is cancelled. This usually occurs due to the underlying Scheduler,
    and is not something we do directly, other than when shutting down the scheduler.

    ```python
    @scheduler.on_future_cancelled
    def my_callback(future: Future):
        ...
    ```
    """
    on_finishing: Subscriber[[]]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when the
    scheduler is finishing up. This occurs right before the scheduler shuts down
    the executor.

    ```python
    @scheduler.on_finishing
    def my_callback():
        ...
    ```
    """
    on_finished: Subscriber[[]]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when
    the scheduler is finished, has shutdown the executor and possibly
    terminated any remaining compute.

    ```python
    @scheduler.on_finished
    def my_callback():
        ...
    ```
    """
    on_stop: Subscriber[str, BaseException | None]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when the
    scheduler is has been stopped due to the [`stop()`][amltk.scheduling.Scheduler.stop]
    method being called.

    ```python
    @scheduler.on_stop
    def my_callback(stop_msg: str, exception: BaseException | None):
        ...
    ```
    """
    on_timeout: Subscriber[[]]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when
    the scheduler reaches the timeout.

    ```python
    @scheduler.on_timeout
    def my_callback():
        ...
    ```
    """
    on_empty: Subscriber[[]]
    """A [`Subscriber`][amltk.scheduling.events.Subscriber] which is called when the
    queue is empty. This can be useful to re-fill the queue and prevent the
    scheduler from exiting.

    ```python
    @scheduler.on_empty
    def my_callback():
        ...
    ```
    """

    STARTED: Event[[]] = Event("on_start")
    FINISHING: Event[[]] = Event("on_finishing")
    FINISHED: Event[[]] = Event("on_finished")
    STOP: Event[str, BaseException | None] = Event("on_stop")
    TIMEOUT: Event[[]] = Event("on_timeout")
    EMPTY: Event[[]] = Event("on_empty")
    FUTURE_SUBMITTED: Event[Future] = Event("on_future_submitted")
    FUTURE_DONE: Event[Future] = Event("on_future_done")
    FUTURE_CANCELLED: Event[Future] = Event("on_future_cancelled")
    FUTURE_RESULT: Event[Future, Any] = Event("on_future_result")
    FUTURE_EXCEPTION: Event[Future, BaseException] = Event("on_future_exception")

    def __init__(
        self,
        executor: Executor,
        *,
        terminate: Callable[[Executor], None] | bool = True,
    ) -> None:
        """Initialize a scheduler.

        Args:
            executor: The dispatcher to use for submitting tasks.
            terminate: Whether to call shutdown on the executor when
                `run(..., wait=False)`. If True, the executor will be
                `shutdown(wait=False)` and we will attempt to terminate
                any workers of the executor. For some `Executors` this
                is enough, i.e. Dask, however for something like
                `ProcessPoolExecutor`, we will use `psutil` to kill
                its worker processes. If a callable, we will use this
                function for custom worker termination.
                If False, shutdown will not be called and the executor will
                remain active.
        """
        super().__init__()
        self.executor = executor
        self.unique_ref = f"Scheduler-{uuid4()}"
        self.emitter = Emitter()
        self.event_counts = self.emitter.event_counts

        # The current state of things and references to them
        self.queue = {}

        # Set up subscribers for events
        self.on_start = self.emitter.subscriber(self.STARTED)
        self.on_finishing = self.emitter.subscriber(self.FINISHING)
        self.on_finished = self.emitter.subscriber(self.FINISHED)
        self.on_stop = self.emitter.subscriber(self.STOP)
        self.on_timeout = self.emitter.subscriber(self.TIMEOUT)
        self.on_empty = self.emitter.subscriber(self.EMPTY)

        self.on_future_submitted = self.emitter.subscriber(self.FUTURE_SUBMITTED)
        self.on_future_done = self.emitter.subscriber(self.FUTURE_DONE)
        self.on_future_cancelled = self.emitter.subscriber(self.FUTURE_CANCELLED)
        self.on_future_exception = self.emitter.subscriber(self.FUTURE_EXCEPTION)
        self.on_future_result = self.emitter.subscriber(self.FUTURE_RESULT)

        self._terminate: Callable[[Executor], None] | None
        if terminate is True:
            self._terminate = termination_strategy(executor)
        else:
            self._terminate = terminate if callable(terminate) else None

        # This can be triggered either by `scheduler.stop` in a callback.
        # Has to be created inside the event loop so there's no issues
        self._stop_event: ContextEvent | None = None

        # This is a condition to make sure monitoring the queue will wait properly
        self._queue_has_items_event = asyncio.Event()

        # This is triggered when run is called
        self._running_event = asyncio.Event()

        # This is set once `run` is called
        self._end_on_exception_flag = Flag(initial=False)

        # This is used to manage suequential queues, where we need a Thread
        # timer to ensure that we don't get caught in an endless loop waiting
        # for the `timeout` in `_run_scheduler` to trigger. This won't trigger
        # because the sync code of submit could possibly keep calling itself
        # endlessly, preventing any of the async code from running.
        self._timeout_timer: Timer | None = None

        # A collection of things that want to register as being part of something
        # to render when the Scheduler is rendered.
        self._renderables: list[RenderableType] = [self.emitter]

        # These are extra user provided renderables during a call to `run()`. We
        # seperate these out so that we can remove them when the scheduler is
        # stopped.
        self._extra_renderables: list[RenderableType] | None = None

        # An indicator an object to render live output (if requested) with
        # `display=` on a call to `run()`
        self._live_output: Live | None = None

    @classmethod
    def with_processes(
        cls,
        max_workers: int | None = None,
        mp_context: BaseContext | Literal["fork", "spawn", "forkserver"] | None = None,
        initializer: Callable[..., Any] | None = None,
        initargs: tuple[Any, ...] = (),
    ) -> Self:
        """Create a scheduler with a `ProcessPoolExecutor`.

        See [`ProcessPoolExecutor`][concurrent.futures.ProcessPoolExecutor]
        for more details.
        """
        if isinstance(mp_context, str):
            from multiprocessing import get_context

            mp_context = get_context(mp_context)

        executor = ProcessPoolExecutor(
            max_workers=max_workers,
            mp_context=mp_context,
            initializer=initializer,
            initargs=initargs,
        )
        return cls(executor=executor)

    @classmethod
    def with_loky(  # noqa: PLR0913
        cls,
        max_workers: int | None = None,
        context: BaseContext | Literal["fork", "spawn", "forkserver"] | None = None,
        timeout: int = 10,
        kill_workers: bool = False,  # noqa: FBT002, FBT001
        reuse: bool | Literal["auto"] = "auto",
        job_reducers: Any | None = None,
        result_reducers: Any | None = None,
        initializer: Callable[..., Any] | None = None,
        initargs: tuple[Any, ...] = (),
        env: dict[str, str] | None = None,
    ) -> Self:
        """Create a scheduler with a `loky.get_reusable_executor`.

        See [loky documentation][https://loky.readthedocs.io/en/stable/API.html]
        for more details.
        """
        from loky import get_reusable_executor

        executor = get_reusable_executor(
            max_workers=max_workers,
            context=context,
            timeout=timeout,
            kill_workers=kill_workers,
            reuse=reuse,  # type: ignore
            job_reducers=job_reducers,
            result_reducers=result_reducers,
            initializer=initializer,
            initargs=initargs,
            env=env,
        )
        return cls(executor=executor)

    @classmethod
    def with_sequential(cls) -> Self:
        """Create a Scheduler that runs sequentially.

        This is useful for debugging and testing. Uses
        a [`SequentialExecutor`][amltk.scheduling.SequentialExecutor].
        """
        return cls(executor=SequentialExecutor())

    @classmethod
    def with_slurm(
        cls,
        *,
        n_workers: int,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler that runs on a SLURM cluster.

        This is useful for running on a SLURM cluster. Uses
        [dask_jobqueue.SLURMCluster][].

        Args:
            n_workers: The number of workers to start.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Returns:
            A scheduler that will run on a SLURM cluster.
        """
        return cls.with_dask_jobqueue(
            "slurm",
            n_workers=n_workers,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )

    @classmethod
    def with_pbs(
        cls,
        *,
        n_workers: int,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler that runs on a PBS cluster.

        This is useful for running on a PBS cluster. Uses
        [dask_jobqueue.PBSCluster][].

        Args:
            n_workers: The number of workers to start.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Returns:
            A scheduler that will run on a PBS cluster.
        """
        return cls.with_dask_jobqueue(
            "pbs",
            n_workers=n_workers,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )

    @classmethod
    def with_sge(
        cls,
        *,
        n_workers: int,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler that runs on a SGE cluster.

        This is useful for running on a SGE cluster. Uses
        [dask_jobqueue.SGECluster][].

        Args:
            n_workers: The number of workers to start.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Returns:
            A scheduler that will run on a SGE cluster.
        """
        return cls.with_dask_jobqueue(
            "sge",
            n_workers=n_workers,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )

    @classmethod
    def with_oar(
        cls,
        *,
        n_workers: int,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler that runs on a OAR cluster.

        This is useful for running on a OAR cluster. Uses
        [dask_jobqueue.OARCluster][].

        Args:
            n_workers: The number of workers to start.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Returns:
            A scheduler that will run on a OAR cluster.
        """
        return cls.with_dask_jobqueue(
            "oar",
            n_workers=n_workers,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )

    @classmethod
    def with_moab(
        cls,
        *,
        n_workers: int,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler that runs on a Moab cluster.

        This is useful for running on a Moab cluster. Uses
        [dask_jobqueue.MoabCluster][].

        Args:
            n_workers: The number of workers to start.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Returns:
            A scheduler that will run on a Moab cluster.
        """
        return cls.with_dask_jobqueue(
            "moab",
            n_workers=n_workers,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )

    @classmethod
    def with_lsf(
        cls,
        *,
        n_workers: int,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler that runs on a LSF cluster.

        This is useful for running on a LSF cluster. Uses
        [dask_jobqueue.LSFCluster][].

        Args:
            n_workers: The number of workers to start.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Returns:
            A scheduler that will run on a LSF cluster.
        """
        return cls.with_dask_jobqueue(
            "lsf",
            n_workers=n_workers,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )

    @classmethod
    def with_htcondor(
        cls,
        *,
        n_workers: int,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler that runs on a HTCondor cluster.

        This is useful for running on a HTCondor cluster. Uses
        [dask_jobqueue.HTCondorCluster][].

        Args:
            n_workers: The number of workers to start.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Returns:
            A scheduler that will run on a HTCondor cluster.
        """
        return cls.with_dask_jobqueue(
            "htcondor",
            n_workers=n_workers,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )

    @classmethod
    def with_dask_jobqueue(
        cls,
        name: DJQ_NAMES,
        *,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        n_workers: int,
        **kwargs: Any,
    ) -> Self:
        """Create a Scheduler with using `dask-jobqueue`.

        See [`dask_jobqueue`][dask_jobqueue] for more details.

        [dask_jobqueue]: https://jobqueue.dask.org/en/latest/

        Args:
            name: The name of the jobqueue to use. This is the name of the
                class in `dask_jobqueue` to use. For example, to use
                `dask_jobqueue.SLURMCluster`, you would use `slurm`.
            n_workers: The number of workers to start.
            submit_command: Overwrite the command to submit a worker if necessary.
            cancel_command: Overwrite the command to cancel a worker if necessary.
            kwargs: Any additional keyword arguments to pass to the
                `dask_jobqueue` class.

        Raises:
            ImportError: If `dask-jobqueue` is not installed.

        Returns:
            A new scheduler with a `dask_jobqueue` executor.
        """
        try:
            from amltk.scheduling.executors.dask_jobqueue import DaskJobqueueExecutor

        except ImportError as e:
            raise ImportError(
                f"To use the {name} executor, you must install the "
                "`dask-jobqueue` package.",
            ) from e

        executor = DaskJobqueueExecutor.from_str(
            name,
            n_workers=n_workers,
            submit_command=submit_command,
            cancel_command=cancel_command,
            **kwargs,
        )
        return cls(executor)

    def empty(self) -> bool:
        """Check if the scheduler is empty.

        Returns:
            True if there are no tasks in the queue.
        """
        return len(self.queue) == 0

    def running(self) -> bool:
        """Whether the scheduler is running and accepting tasks to dispatch.

        Returns:
            True if the scheduler is running and accepting tasks.
        """
        return self._running_event.is_set()

    def submit(
        self,
        fn: Callable[P, R],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Future[R]:
        """Submits a callable to be executed with the given arguments.

        Args:
            fn: The callable to be executed as
                fn(*args, **kwargs) that returns a Future instance representing
                the execution of the callable.
            args: positional arguments to pass to the function
            kwargs: keyword arguments to pass to the function

        Raises:
            SchedulerNotRunningError: If the scheduler is not running.
                You can protect against this using,
                [`scheduler.running()`][amltk.scheduling.scheduler.Scheduler.running].

        Returns:
            A Future representing the given call.
        """
        if not self.running():
            msg = (
                f"Scheduler is not running, cannot submit task {fn}"
                f" with {args=}, {kwargs=}"
            )
            raise SchedulerNotRunningError(msg)

        try:
            sync_future = self.executor.submit(fn, *args, **kwargs)
            future = asyncio.wrap_future(sync_future)
        except Exception as e:
            logger.exception(f"Could not submit task {fn}", exc_info=e)
            raise e

        self._register_future(future, fn, *args, **kwargs)
        return future

    @overload
    def task(
        self,
        function: Callable[Concatenate[Comm, P], R],
        *,
        plugins: Comm.Plugin | Iterable[Comm.Plugin | Plugin] = ...,
        init_plugins: bool = ...,
    ) -> Task[P, R]:
        ...

    @overload
    def task(
        self,
        function: Callable[P, R],
        *,
        plugins: Plugin | Iterable[Plugin] = (),
        init_plugins: bool = True,
    ) -> Task[P, R]:
        ...

    def task(
        self,
        function: Callable[P, R] | Callable[Concatenate[Comm, P], R],
        *,
        plugins: Plugin | Iterable[Plugin] = (),
        init_plugins: bool = True,
    ) -> Task[P, R]:
        """Create a new task.

        Args:
            function: The function to run using the scheduler.
            plugins: The plugins to attach to the task.
            init_plugins: Whether to initialize the plugins.

        Returns:
            A new task.
        """
        # HACK: Not that the type: ignore is due to the fact that we can't use type
        # checking to enforce that
        # A. `function` is a callable with the first arg being a Comm
        # B. `plugins`
        task = Task(function, self, plugins=plugins, init_plugins=init_plugins)  # type: ignore
        self.add_renderable(task)
        return task  # type: ignore

    def _register_future(
        self,
        future: Future,
        function: Callable,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Registers the future into the queue and add a callback that will be called
        upon future completion. This callback will remove the future from the queue.
        """
        self.queue[future] = (function, args, kwargs)
        self._queue_has_items_event.set()

        self.on_future_submitted.emit(future)
        # Display if requested
        if self._live_output:
            self._live_output.refresh()
            future.add_done_callback(
                lambda _, live=self._live_output: live.refresh(),  # type: ignore
            )

        future.add_done_callback(self._register_complete)

    def _register_complete(self, future: Future) -> None:
        try:
            self.queue.pop(future)
        except ValueError as e:
            logger.error(
                f"{future=} was not found in the queue {self.queue}: {e}!",
                exc_info=True,
            )

        if future.cancelled():
            self.on_future_cancelled.emit(future)
            return

        self.on_future_done.emit(future)

        exception = future.exception()
        if exception:
            self.on_future_exception.emit(future, exception)
            if self._end_on_exception_flag and future.done():
                self.stop(stop_msg="Ending on first exception", exception=exception)
        else:
            result = future.result()
            self.on_future_result.emit(future, result)

    async def _monitor_queue_empty(self) -> None:
        """Monitor for the queue being empty and trigger an event when it is."""
        if not self.running():
            raise RuntimeError("The scheduler is not running!")

        while True:
            while self.queue:
                queue = list(self.queue)
                await asyncio.wait(queue, return_when=asyncio.ALL_COMPLETED)

            # Signal that the queue is now empty
            self._queue_has_items_event.clear()
            self.on_empty.emit()

            # Wait for an item to be in the queue
            await self._queue_has_items_event.wait()

            logger.debug("Queue has been filled again")

    async def _stop_when_triggered(self, stop_event: ContextEvent) -> bool:
        """Stop the scheduler when the stop event is set."""
        if not self.running():
            raise RuntimeError("The scheduler is not running!")

        await stop_event.wait()

        logger.debug("Stop event triggered, stopping scheduler")
        return True

    async def _run_scheduler(  # noqa: C901, PLR0912, PLR0915
        self,
        *,
        timeout: float | None = None,
        end_on_empty: bool = True,
        wait: bool = True,
    ) -> ExitState.Code | BaseException:
        self.executor.__enter__()
        self._stop_event = ContextEvent()

        if self._live_output is not None:
            self._live_output.__enter__()

            # If we are doing a live display, we have to disable
            # warnings as they will screw up the display rendering
            # However, we re-enable it after the scheduler has finished running
            warning_catcher = warnings.catch_warnings()
            warning_catcher.__enter__()
            warnings.filterwarnings("ignore")
        else:
            warning_catcher = None

        # Declare we are running
        self._running_event.set()

        # Start a Thread Timer as our timing mechanism.
        # HACK: This is required because the SequentialExecutor mode
        # will not allow the async loop to run, meaning we can't update
        # any internal state.
        if timeout is not None:
            self._timeout_timer = Timer(timeout, lambda: None)
            self._timeout_timer.start()

        self.on_start.emit()

        # Monitor for `stop` being triggered
        stop_triggered = asyncio.create_task(
            self._stop_when_triggered(self._stop_event),
        )

        # Monitor for the queue being empty
        monitor_empty = asyncio.create_task(self._monitor_queue_empty())
        if end_on_empty:
            self.on_empty(lambda: monitor_empty.cancel(), hidden=True)

        # The timeout criterion is satisfied by the `timeout` arg
        await asyncio.wait(
            [stop_triggered, monitor_empty],
            timeout=timeout,
            return_when=asyncio.FIRST_COMPLETED,
        )

        # Determine the reason for stopping
        stop_reason: BaseException | ExitState.Code
        if stop_triggered.done() and self._stop_event.is_set():
            stop_reason = ExitState.Code.STOPPED

            msg, exception = self._stop_event.context
            _log = logger.exception if exception else logger.debug
            _log(f"Stop Message: {msg}", exc_info=exception)

            self.on_stop.emit(str(msg), exception)
            if self._end_on_exception_flag and exception:
                stop_reason = exception
            else:
                stop_reason = ExitState.Code.STOPPED
        elif monitor_empty.done():
            logger.debug("Scheduler stopped due to being empty.")
            stop_reason = ExitState.Code.EXHAUSTED
        elif timeout is not None:
            logger.debug(f"Scheduler stopping as {timeout=} reached.")
            stop_reason = ExitState.Code.TIMEOUT
            self.on_timeout.emit()
        else:
            logger.warning("Scheduler stopping for unknown reason!")
            stop_reason = ExitState.Code.UNKNOWN

        # Stop all running async tasks, i.e. monitoring the queue to trigger an event
        tasks = [monitor_empty, stop_triggered]
        for task in tasks:
            task.cancel()

        # Await all the cancelled tasks and read the exceptions
        await asyncio.gather(*tasks, return_exceptions=True)

        self.on_finishing.emit()
        logger.debug("Scheduler is finished")
        logger.debug(f"Shutting down scheduler executor with {wait=}")

        # The scheduler is now refusing jobs
        self._running_event.clear()
        logger.debug("Scheduler has shutdown and declared as no longer running")

        # This will try to end the tasks based on wait and self._terminate
        Scheduler._end_pending(
            wait=wait,
            futures=list(self.queue.keys()),
            executor=self.executor,
            termination_strategy=self._terminate,
        )

        self.on_finished.emit()
        logger.debug(f"Scheduler finished with status {stop_reason}")

        # Clear all events
        self._stop_event.clear()
        self._queue_has_items_event.clear()

        if self._live_output is not None:
            self._live_output.refresh()
            self._live_output.stop()

        if self._timeout_timer is not None:
            self._timeout_timer.cancel()

        if warning_catcher is not None:
            warning_catcher.__exit__()  # type: ignore

        return stop_reason

    def run(
        self,
        *,
        timeout: float | None = None,
        end_on_empty: bool = True,
        wait: bool = True,
        on_exception: Literal["raise", "end", "ignore"] = "raise",
        asyncio_debug_mode: bool = False,
        display: bool | list[RenderableType] = False,
    ) -> ExitState:
        """Run the scheduler.

        Args:
            timeout: The maximum time to run the scheduler for in
                seconds. Defaults to `None` which means no timeout and it
                will end once the queue is empty if `end_on_empty=True`.
            end_on_empty: Whether to end the scheduler when the queue becomes empty.
            wait: Whether to wait for currently running compute to finish once
                the `Scheduler` is shutting down.

                * If `#!python True`, will wait for all currently running compute.
                * If `#!python False`, will attempt to cancel/terminate all currently
                    running compute and shutdown the executor. This may be useful
                    if you want to end the scheduler as quickly as possible or
                    respect the `timeout=` more precisely.
            on_exception: What to do when an exception occurs in the scheduler
                or callbacks (**Does not apply to submitted compute!**)

                * If `#!python "raise"`, the scheduler will stop and raise the
                    exception at the point where you called `run()`.
                * If `#!python "ignore"`, the scheduler will continue running,
                    ignoring the exception. This may be useful when requiring more
                    robust execution.
                * If `#!python "end"`, similar to `#!python "raise"`, the scheduler
                    will stop but no exception will occur and the control flow
                    will return gracefully to the point where you called `run()`.
            asyncio_debug_mode: Whether to run the async loop in debug mode.
                Defaults to `False`. Please see [asyncio.run][] for more.
            display: Whether to display the scheduler live in the console.

                * If `#!python True`, will display the scheduler and all its tasks.
                * If a `#!python list[RenderableType]` , will display the scheduler
                    itself plus those renderables.

        Returns:
            The reason for the scheduler ending.

        Raises:
            RuntimeError: If the scheduler is already running.
        """
        return asyncio.run(
            self.async_run(
                timeout=timeout,
                end_on_empty=end_on_empty,
                wait=wait,
                on_exception=on_exception,
                display=display,
            ),
            debug=asyncio_debug_mode,
        )

    async def async_run(
        self,
        *,
        timeout: float | None = None,
        end_on_empty: bool = True,
        wait: bool = True,
        on_exception: Literal["raise", "end", "ignore"] = "raise",
        display: bool | list[RenderableType] = False,
    ) -> ExitState:
        """Async version of `run`.

        This can be useful if you are already running in an async context,
        such as in a web server or Jupyter notebook.

        Please see [`run()`][amltk.Scheduler.run] for more details.
        """
        if self.running():
            raise RuntimeError("Scheduler already seems to be running")

        logger.debug("Starting scheduler")

        # Make sure flags are set
        self._end_on_exception_flag.set(value=on_exception in ("raise", "end"))

        # If the user has requested to have a live display,
        # we will need to setup a `Live` instance to render to
        if display:
            from rich.live import Live

            if isinstance(display, list):
                self._extra_renderables = display

            self._live_output = Live(
                auto_refresh=False,
                get_renderable=self.__rich__,
            )

        loop = asyncio.get_running_loop()

        # Set the exception handler for asyncio
        previous_exception_handler = None
        if on_exception in ("raise", "end"):
            previous_exception_handler = loop.get_exception_handler()

            def custom_exception_handler(
                loop: asyncio.AbstractEventLoop,
                context: dict[str, Any],
            ) -> None:
                exception = context.get("exception")
                message = context.get("message")

                # handle with previous handler
                if previous_exception_handler:
                    previous_exception_handler(loop, context)
                else:
                    loop.default_exception_handler(context)

                self.stop(stop_msg=message, exception=exception)

            loop.set_exception_handler(custom_exception_handler)

        # Run the actual scheduling loop
        result = await self._run_scheduler(
            timeout=timeout,
            end_on_empty=end_on_empty,
            wait=wait,
        )

        # Reset variables back to its default
        self._live_output = None
        self._extra_renderables = None
        self._end_on_exception_flag.reset()

        if previous_exception_handler is not None:
            loop.set_exception_handler(previous_exception_handler)

        # If we were meant to end on an exception and the result
        # we got back from the scheduler was an exception, raise it
        if isinstance(result, BaseException):
            if on_exception == "raise":
                raise result

            return ExitState(code=ExitState.Code.EXCEPTION, exception=result)

        return ExitState(code=result)

    def stop(
        self,
        *args: Any,
        stop_msg: str | None = None,
        exception: BaseException | None = None,
        **kwargs: Any,
    ) -> None:
        """Stop the scheduler.

        The scheduler will stop, finishing currently running tasks depending
        on the `wait=` parameter to [`Scheduler.run`][amltk.Scheduler.run].

        The call signature is kept open with `*args, **kwargs` to make it
        easier to include in any callback.

        Args:
            *args: Logged in a debug message
            **kwargs: Logged in a debug message
            stop_msg: The message to log when stopping the scheduler.
            exception: The exception which incited `stop()` to be called.
                Will be used by the `Scheduler` to possibly raise the exception
                to the user.
        """
        if not self.running():
            return

        assert self._stop_event is not None

        msg = stop_msg if stop_msg is not None else "scheduler.stop() was called."
        logger.debug(f"Stopping scheduler: {msg} {args=} {kwargs=}")

        self._stop_event.set(msg=msg, exception=exception)
        self._running_event.clear()

    def call_later(
        self,
        delay: float,
        fn: Callable[P, Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> asyncio.TimerHandle:
        """Schedule a function to be run after a delay.

        Args:
            delay: The delay in seconds.
            fn: The function to run.
            args: The positional arguments to pass to the function.
            kwargs: The keyword arguments to pass to the function.

        Returns:
            A timer handle that can be used to cancel the function.
        """
        if not self.running():
            raise RuntimeError("Scheduler is not running!")

        _fn = partial(fn, *args, **kwargs)
        loop = asyncio.get_running_loop()
        return loop.call_later(delay, _fn)

    @staticmethod
    def _end_pending(
        *,
        futures: list[Future],
        executor: Executor,
        wait: bool = True,
        termination_strategy: Callable[[Executor], Any] | None = None,
    ) -> None:
        if wait:
            logger.debug("Waiting for currently running tasks to finish.")
            executor.shutdown(wait=wait)
        elif termination_strategy is None:
            logger.warning(
                "Cancelling currently running tasks and then waiting "
                f" as there is no termination strategy provided for {executor=}`.",
            )
            # Just try to cancel the tasks. Will cancel pending tasks
            # but executors like dask will even kill the job
            for future in futures:
                if not future.done():
                    logger.debug(f"Cancelling {future=}")
                    future.cancel()

            # Here we wait, if we could  cancel, then we wait for that
            # to happen, otherwise we are just waiting as anticipated.
            executor.shutdown(wait=wait)
        else:
            logger.debug(f"Terminating workers with {termination_strategy=}")
            for future in futures:
                if not future.done():
                    logger.debug(f"Cancelling {future=}")
                    future.cancel()
            termination_strategy(executor)
            executor.shutdown(wait=wait)

    def add_renderable(self, renderable: RenderableType) -> None:
        """Add a renderable object to the scheduler.

        This will be displayed whenever the scheduler is displayed.
        """
        self._renderables.append(renderable)

    def __rich__(self) -> RenderableType:
        from rich.console import Group
        from rich.panel import Panel
        from rich.table import Column, Table
        from rich.text import Text
        from rich.tree import Tree

        from amltk._richutil import richify
        from amltk._richutil.renderers.function import Function

        MAX_FUTURE_ITEMS = 5
        OFFSETS = 1 + 1 + 2  # Header + ellipses space + panel borders

        title = Text("Scheduler", style="magenta bold")
        if self.running():
            title.append(" (running)", style="green")

        future_table = Table.grid()

        # Select the most latest items
        future_items = list(self.queue.items())[-MAX_FUTURE_ITEMS:]
        for future, (func, args, kwargs) in future_items:
            entry = Function(
                func,
                (args, kwargs),
                link=False,
                prefix=future._state,
                no_wrap=True,
            )
            future_table.add_row(entry)

        if len(self.queue) > MAX_FUTURE_ITEMS:
            future_table.add_row(Text("...", style="yellow"))

        queue_column_text = Text.assemble(
            "Queue: (",
            (f"{len(self.queue)}", "yellow"),
            ")",
        )

        layout_table = Table(
            Column("Executor", ratio=1),
            Column(queue_column_text, ratio=2),
            box=None,
            expand=True,
            padding=(0, 1),
        )
        layout_table.add_row(richify(self.executor), future_table)

        panel = Panel(
            layout_table,
            title=title,
            title_align="left",
            border_style="magenta",
            height=MAX_FUTURE_ITEMS + OFFSETS,
        )
        tree = Tree(panel, guide_style="magenta bold")

        for renderable in self._renderables:
            tree.add(renderable)

        if not self._extra_renderables:
            return tree

        return Group(tree, *self._extra_renderables)


@dataclass
class ExitState:
    """The exit state of a scheduler.

    Attributes:
        reason: The reason for the exit.
        exception: The exception that caused the exit, if any.
    """

    code: ExitState.Code
    exception: BaseException | None = None

    class Code(Enum):
        """The reason the scheduler ended."""

        STOPPED = auto()
        """The scheduler was stopped forcefully with `Scheduler.stop`."""

        TIMEOUT = auto()
        """The scheduler finished because of a timeout."""

        EXHAUSTED = auto()
        """The scheduler finished because it exhausted its queue."""

        CANCELLED = auto()
        """The scheduler was cancelled."""

        UNKNOWN = auto()
        """The scheduler finished for an unknown reason."""

        EXCEPTION = auto()
        """The scheduler finished because of an exception."""



from amltk.data.conversions import flatten_if_1d, probabilities_to_classes, to_numpy
from amltk.data.dtype_reduction import (
    reduce_dtypes,
    reduce_floating_precision,
    reduce_int_span,
)
from amltk.data.measure import byte_size

__all__ = [
    "byte_size",
    "reduce_dtypes",
    "reduce_floating_precision",
    "reduce_int_span",
    "probabilities_to_classes",
    "to_numpy",
    "flatten_if_1d",
]



"""The base [`Optimizer`][amltk.optimization.optimizer.Optimizer] class,
defines the API we require optimizers to implement.

* [`ask()`][amltk.optimization.optimizer.Optimizer.ask] - Ask the optimizer for a
    new [`Trial`][amltk.optimization.trial.Trial] to evaluate.
* [`tell()`][amltk.optimization.optimizer.Optimizer.tell] - Tell the optimizer
    the result of the sampled config. This comes in the form of a
    [`Trial.Report`][amltk.optimization.trial.Trial.Report].

Additionally, to aid users from switching between optimizers, the
[`preferred_parser()`][amltk.optimization.optimizer.Optimizer.preferred_parser]
method should return either a `parser` function or a string that can be used
with [`node.search_space(parser=..._)`][amltk.pipeline.Node.search_space] to
extract the search space for the optimizer.
"""
from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable, Sequence
from datetime import datetime
from typing import TYPE_CHECKING, Any, Concatenate, Generic, ParamSpec, TypeVar

from amltk.store.paths.path_bucket import PathBucket

if TYPE_CHECKING:
    from amltk.optimization.metric import Metric
    from amltk.optimization.trial import Trial
    from amltk.pipeline import Node

I = TypeVar("I")  # noqa: E741
P = ParamSpec("P")
ParserOutput = TypeVar("ParserOutput")


class Optimizer(Generic[I]):
    """An optimizer protocol.

    An optimizer is an object that can be asked for a trail using `ask` and a
    `tell` to inform the optimizer of the report from that trial.
    """

    metrics: Sequence[Metric]
    """The metrics to optimize."""

    bucket: PathBucket
    """The bucket to give to trials generated by this optimizer."""

    def __init__(
        self,
        metrics: Sequence[Metric],
        bucket: PathBucket | None = None,
    ) -> None:
        """Initialize the optimizer.

        Args:
            metrics: The metrics to optimize.
            bucket: The bucket to store results of individual trials from this
                optimizer.
        """
        super().__init__()
        self.metrics = metrics
        self.bucket = (
            bucket
            if bucket is not None
            else PathBucket(f"{self.__class__.__name__}-{datetime.now().isoformat()}")
        )

    @abstractmethod
    def tell(self, report: Trial.Report[I]) -> None:
        """Tell the optimizer the report for an asked trial.

        Args:
            report: The report for a trial
        """

    @abstractmethod
    def ask(self) -> Trial[I]:
        """Ask the optimizer for a trial to evaluate.

        Returns:
            A config to sample.
        """
        ...

    @classmethod
    def preferred_parser(
        cls,
    ) -> str | Callable[Concatenate[Node, ...], Any] | Callable[[Node], Any] | None:
        """The preferred parser for this optimizer.

        !!! note

            Subclasses should override this as required.

        """
        return None



"""The [`Comm.Plugin`][amltk.scheduling.plugins.comm.Comm.Plugin] enables
two way-communication with running [`Task`][amltk.scheduling.task.Task].

The [`Comm`][amltk.scheduling.plugins.comm.Comm] provides an easy interface to
communicate while the [`Comm.Msg`][amltk.scheduling.plugins.comm.Comm.Msg] encapsulates
messages between the main process and the `Task`.

??? tip "Usage"

    To setup a `Task` to work with a `Comm`, the `Task` **must accept a `comm` as
    it's first argument**.

    ```python exec="true" source="material-block" result="python" hl_lines="4-7 10 17-19 21-23"
    from amltk.scheduling import Scheduler
    from amltk.scheduling.plugins import Comm

    def powers_of_two(comm: Comm, start: int, n: int) -> None:
        with comm.open():
            for i in range(n):
                comm.send(start ** (i+1))
    from amltk._doc import make_picklable; make_picklable(powers_of_two)  # markdown-exec: hide

    scheduler = Scheduler.with_processes(1)
    task = scheduler.task(powers_of_two, plugins=Comm.Plugin())
    results = []

    @scheduler.on_start
    def on_start():
        task.submit(2, 5)

    @task.on("comm-open")
    def on_open(msg: Comm.Msg):
        print(f"Task has opened | {msg}")

    @task.on("comm-message")
    def on_message(msg: Comm.Msg):
        results.append(msg.data)

    scheduler.run()
    print(results)
    ```

    You can also block a worker, waiting for a response from the main process, allowing for the
    worker to [`request()`][amltk.scheduling.plugins.comm.Comm.request] data from the main process.

    ```python exec="true" source="material-block" result="python" hl_lines="7 20-23"
    from amltk.scheduling import Scheduler
    from amltk.scheduling.plugins import Comm

    def my_worker(comm: Comm, n_tasks: int) -> None:
        with comm.open():
            for task_number in range(n_tasks):
                task = comm.request(task_number)
                comm.send(f"Task recieved {task} for {task_number}")
    from amltk._doc import make_picklable; make_picklable(my_worker)  # markdown-exec: hide

    scheduler = Scheduler.with_processes(1)
    task = scheduler.task(my_worker, plugins=Comm.Plugin())

    items = ["A", "B", "C"]
    results = []

    @scheduler.on_start
    def on_start():
        task.submit(n_tasks=3)

    @task.on("comm-request")
    def on_request(msg: Comm.Msg):
        task_number = msg.data
        msg.respond(items[task_number])

    @task.on("comm-message")
    def on_message(msg: Comm.Msg):
        results.append(msg.data)

    scheduler.run()
    print(results)
    ```

??? example "`@events`"

    === "`@comm-message`"

        ::: amltk.scheduling.plugins.comm.Comm.MESSAGE

    === "`@comm-request`"

        ::: amltk.scheduling.plugins.comm.Comm.REQUEST

    === "`@comm-open`"

        ::: amltk.scheduling.plugins.comm.Comm.OPEN

    === "`@comm-close`"

        ::: amltk.scheduling.plugins.comm.Comm.CLOSE

??? warning "Supported Backends"

    The current implementation relies on [`Pipe`][multiprocessing.Pipe] which only
    works between processes on the same system/cluster. There is also limited support
    with `dask` backends.

    This could be extended to allow for web sockets or other forms of connections
    but requires time. Please let us know in the Github issues if this is something
    you are interested in!
"""  # noqa: E501
from __future__ import annotations

import asyncio
import logging
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from dataclasses import dataclass, field
from enum import Enum
from multiprocessing import (
    Pipe,
    TimeoutError as MPTimeoutError,
)
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    TypeAlias,
    TypeVar,
)
from typing_extensions import ParamSpec, override

from amltk._asyncm import AsyncConnection
from amltk.scheduling.events import Event
from amltk.scheduling.plugins.plugin import Plugin as TaskPlugin

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from multiprocessing.connection import Connection
    from typing_extensions import Self

    from rich.panel import Panel

    from amltk.scheduling.task import Task

    CommID: TypeAlias = int


T = TypeVar("T")
M = TypeVar("M")
P = ParamSpec("P")
R = TypeVar("R")


@dataclass
class AsyncComm:
    """A async wrapper of a Comm."""

    comm: Comm

    async def request(
        self,
        *,
        timeout: float | None = None,
    ) -> Any:
        """Recieve a message.

        Args:
            timeout: The timeout in seconds to wait for a message, raises
                a [`Comm.TimeoutError`][amltk.scheduling.plugins.comm.Comm.TimeoutError]
                if the timeout is reached.
                If `None`, will wait forever.

        Returns:
            The message from the worker or the default value.
        """
        connection = AsyncConnection(self.comm.connection)
        try:
            return await asyncio.wait_for(connection.recv(), timeout=timeout)
        except asyncio.TimeoutError as e:
            raise Comm.TimeoutError(
                f"Timed out waiting for response from {self.comm}",
            ) from e

    async def send(self, obj: Any) -> None:
        """Send a message.

        Args:
            obj: The message to send.
        """
        return await AsyncConnection(self.comm.connection).send(obj)


class Comm:
    """A communication channel between a worker and scheduler.

    For duplex connections, such as returned by python's builtin
    [`Pipe`][multiprocessing.Pipe], use the
    [`create(duplex=...)`][amltk.Comm.create] class method.

    Adds three new events to the task:

    * [`@comm-message`][amltk.scheduling.plugins.comm.Comm.MESSAGE]
    * [`@comm-request`][amltk.scheduling.plugins.comm.Comm.REQUEST]
    * [`@comm-close`][amltk.scheduling.plugins.comm.Comm.CLOSE]
    * [`@comm-open`][amltk.scheduling.plugins.comm.Comm.OPEN]

    Attributes:
        connection: The underlying Connection
        id: The id of the comm.
    """

    MESSAGE: Event[Comm.Msg] = Event("comm-message")
    """A Task has sent a message to the main process.

    ```python exec="true" source="material-block" html="true" hl_lines="6 11-13"
    from amltk.scheduling import Scheduler
    from amltk.scheduling.plugins import Comm

    def fn(comm: Comm, x: int) -> int:
        with comm.open():
            comm.send(x + 1)

    scheduler = Scheduler.with_processes(1)
    task = scheduler.task(fn, plugins=Comm.Plugin())

    @task.on("comm-message")
    def callback(msg: Comm.Msg):
        print(msg.data)
    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```
    """

    REQUEST: Event[Comm.Msg] = Event("comm-request")
    """A Task has sent a request.

    ```python exec="true" source="material-block" html="true" hl_lines="6 16-18"
    from amltk.scheduling import Scheduler
    from amltk.scheduling.plugins import Comm

    def greeter(comm: Comm, greeting: str) -> None:
        with comm.open():
            name = comm.request()
            comm.send(f"{greeting} {name}!")
    from amltk._doc import make_picklable; make_picklable(greeter)  # markdown-exec: hide

    scheduler = Scheduler.with_processes(1)
    task = scheduler.task(greeter, plugins=Comm.Plugin())

    @scheduler.on_start
    def on_start():
        task.submit("Hello")

    @task.on("comm-request")
    def on_request(msg: Comm.Msg):
        msg.respond("Alice")

    @task.on("comm-message")
    def on_msg(msg: Comm.Msg):
        print(msg.data)

    scheduler.run()
    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```
    """  # noqa: E501

    OPEN: Event[Comm.Msg] = Event("comm-open")
    """The task has signalled it's open.

    ```python exec="true" source="material-block" html="true" hl_lines="5 15-17"
    from amltk.scheduling import Scheduler
    from amltk.scheduling.plugins import Comm

    def fn(comm: Comm) -> None:
        with comm.open():
            pass
   from amltk._doc import make_picklable; make_picklable(fn)  # markdown-exec: hide

    scheduler = Scheduler.with_processes(1)
    task = scheduler.task(fn, plugins=Comm.Plugin())

    @scheduler.on_start
    def on_start():
        task.submit()

    @task.on("comm-open")
    def callback(msg: Comm.Msg):
        print("Comm has just used comm.open()")

    scheduler.run()
    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```
    """

    CLOSE: Event[Comm.Msg] = Event("comm-close")
    """The task has signalled it's close.

    ```python exec="true" source="material-block" html="true" hl_lines="7 17-19"
    from amltk.scheduling import Scheduler
    from amltk.scheduling.plugins import Comm

    def fn(comm: Comm) -> None:
        with comm.open():
            pass
            # Will send a close signal to the main process as it exists this block

        print("Done")
    from amltk._doc import make_picklable; make_picklable(fn)  # markdown-exec: hide
    scheduler = Scheduler.with_processes(1)
    task = scheduler.task(fn, plugins=Comm.Plugin())

    @scheduler.on_start
    def on_start():
        task.submit()

    @task.on("comm-close")
    def on_close(msg: Comm.msg):
        print(f"Worker close with {msg}")

    scheduler.run()
    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```
    """

    def __init__(self, connection: Connection) -> None:
        """Initialize the Comm.

        Args:
            connection: The underlying Connection
        """
        super().__init__()
        self.connection = connection
        self.id: CommID = id(self)

    def _send_pipe(self, obj: Any) -> None:
        self.connection.send(obj)

    def send(self, obj: Any) -> None:
        """Send a message.

        Args:
            obj: The object to send.
        """
        self._send_pipe((Comm.Msg.Kind.MESSAGE, obj))

    def close(  # noqa: PLR0912, C901
        self,
        msg: Any | None = None,
        *,
        wait_for_ack: bool = False,
        okay_if_broken_pipe: bool = False,
        side: str = "",
    ) -> None:
        """Close the connection.

        Args:
            msg: The message to send to the other end of the connection.
            wait_for_ack: If `True`, wait for an acknowledgement from the
                other end before closing the connection.
            okay_if_broken_pipe: If `True`, will not log an error if the
                connection is already closed.
            side: The side of the connection for naming purposes.
        """
        if not self.connection.closed:
            kind = Comm.Msg.Kind.CLOSE_WITH_ACK if wait_for_ack else Comm.Msg.Kind.CLOSE
            try:
                self._send_pipe((kind, msg))
            except BrokenPipeError as e:
                if not okay_if_broken_pipe:
                    logger.error(f"{side} - Error sending close signal: {type(e)}{e}")
            except Exception as e:  # noqa: BLE001
                logger.error(f"{side} - Error sending close signal: {type(e)}{e}")
            else:
                if wait_for_ack:
                    logger.debug(f"{side} - Waiting for ACK")
                    try:
                        recieved_msg = self.connection.recv()
                    except Exception as e:  # noqa: BLE001
                        logger.error(
                            f"{side} - Error waiting for ACK, closing: {type(e)}{e}",
                        )
                    else:
                        match recieved_msg:
                            case Comm.Msg.Kind.WORKER_CLOSE_REQUEST:
                                logger.error(
                                    f"{side} - Worker recieved request to close!",
                                )
                            case Comm.Msg.Kind.ACK:
                                logger.debug(f"{side} - Recieved ACK, closing")
                            case _:
                                logger.warning(
                                    f"{side} - Expected ACK but {recieved_msg=}",
                                )
            finally:
                try:
                    self.connection.close()
                except OSError:
                    # It's possble that the connection was closed by the other end
                    # before we could close it.
                    pass
                except Exception as e:  # noqa: BLE001
                    logger.error(f"{side} - Error closing connection: {type(e)}{e}")

    @classmethod
    def create(cls, *, duplex: bool = True) -> tuple[Self, Self]:
        """Create a pair of communication channels.

        Wraps the output of
        [`multiprocessing.Pipe(duplex=duplex)`][multiprocessing.Pipe].

        Args:
            duplex: Whether to allow for two-way communication

        Returns:
            A pair of communication channels.
        """
        reader, writer = Pipe(duplex=duplex)
        return cls(reader), cls(writer)

    @property
    def as_async(self) -> AsyncComm:
        """Return an async version of this comm."""
        return AsyncComm(self)

    def request(
        self,
        msg: Any | None = None,
        *,
        timeout: None | float = None,
    ) -> Any:
        """Receive a message.

        Args:
            msg: The message to send to the other end of the connection.
                If left empty, will be `None`.
            timeout: If float, will wait for that many seconds, raising an exception
                if exceeded. Otherwise, None will wait forever.

        Raises:
            Comm.TimeoutError: If the timeout is reached.
            Comm.CloseRequestError: If the other end needs to abruptly end and
                can not fufill the request. If thise error is thrown, the worker
                should finish as soon as possible.

        Returns:
            The received message or the default.
        """
        self._send_pipe((Comm.Msg.Kind.REQUEST, msg))
        if not self.connection.poll(timeout):
            raise Comm.TimeoutError(f"Timed out waiting for response for {msg}")

        response = self.connection.recv()
        if response == Comm.Msg.Kind.WORKER_CLOSE_REQUEST:
            logger.error("Worker recieved request to close!")
            raise Comm.CloseRequestError()

        return response

    @contextmanager
    def open(
        self,
        opening_msg: Any | None = None,
        *,
        wait_for_ack: bool = False,
        side: str = "worker",
    ) -> Iterator[Self]:
        """Open the connection.

        Args:
            opening_msg: The message to send to the main process
                when the connection is opened.
            wait_for_ack: If `True`, wait for an acknowledgement from the
                other end before closing the connection and exiting the
                context manager.
            side: The side of the connection for naming purposes.
                Usually this is only done on the `"worker"` side.

        Yields:
            The comm.
        """
        self._send_pipe((Comm.Msg.Kind.OPEN, opening_msg))
        yield self
        self.close(wait_for_ack=wait_for_ack, side=side)

    class Plugin(TaskPlugin):
        """A plugin that handles communication with a worker."""

        name: ClassVar[str] = "comm-plugin"

        def __init__(
            self,
            create_comms: Callable[[], tuple[Comm, Comm]] | None = None,
        ) -> None:
            """Initialize the plugin.

            Args:
                create_comms: A function that creates a pair of communication
                    channels. Defaults to `Comm.create`.
            """
            super().__init__()
            if create_comms is None:
                create_comms = Comm.create

            self.create_comms = create_comms
            self.comms: dict[CommID, tuple[Comm, Comm]] = {}
            self.communication_tasks: list[asyncio.Task] = []
            self.task: Task
            self.open_comms: set[CommID] = set()

        @override
        def attach_task(self, task: Task) -> None:
            """Attach the plugin to a task.

            This method is called when the plugin is attached to a task. This
            is the place to subscribe to events on the task, create new subscribers
            for people to use or even store a reference to the task for later use.

            Args:
                task: The task the plugin is being attached to.
            """
            self.task = task
            task.emitter.add_event(Comm.MESSAGE, Comm.REQUEST, Comm.OPEN, Comm.CLOSE)
            task.on_submitted(self._begin_listening, hidden=True)

        @override
        def pre_submit(
            self,
            fn: Callable[P, R],
            *args: P.args,
            **kwargs: P.kwargs,
        ) -> tuple[Callable[P, R], tuple, dict] | None:
            """Pre-submit hook.

            This method is called before the task is submitted.

            Args:
                fn: The task function.
                *args: The arguments to the task function.
                **kwargs: The keyword arguments to the task function.

            Returns:
                A tuple of the task function, arguments and keyword arguments
                if the task should be submitted, or `None` if the task should
                not be submitted.
            """
            host_comm, worker_comm = self.create_comms()

            # We don't necessarily know if the future will be submitted. If so,
            # we will use this index later to retrieve the host_comm
            self.comms[worker_comm.id] = (host_comm, worker_comm)

            # Make sure to include the Comm
            return fn, (worker_comm, *args), kwargs

        @override
        def copy(self) -> Self:
            """Return a copy of the plugin.

            Please see [`Plugin.copy()`][amltk.scheduling.Plugin.copy].
            """
            return self.__class__(create_comms=self.create_comms)

        def _begin_listening(self, f: asyncio.Future, *args: Any, **_: Any) -> Any:
            match args:
                case (worker_comm, *_) if isinstance(worker_comm, Comm):
                    worker_comm = args[0]
                case _:
                    raise ValueError(f"Expected first arg to be a Comm, got {args[0]}")

            host_comm, worker_comm = self.comms[worker_comm.id]

            coroutine = asyncio.create_task(
                self._communicate(f, host_comm, worker_comm),
            )
            coroutine.add_done_callback(self._deregister_comm_coroutine)

            # NOTE: Asyncio coroutines must have a reference stored somewhere so
            # we need to hold on to it until it's done.
            self.communication_tasks.append(coroutine)

        def _deregister_comm_coroutine(self, coroutine: asyncio.Task) -> None:
            if coroutine in self.communication_tasks:
                self.communication_tasks.remove(coroutine)
            else:
                logger.warning(f"Communication coroutine {coroutine} not found!")

            if (exception := coroutine.exception()) is not None:
                raise exception

        async def _communicate(
            self,
            future: asyncio.Future,
            host_comm: Comm,
            worker_comm: Comm,
        ) -> None:
            """Communicate with the task.

            This is a coroutine that will run until the scheduler is stopped or
            the comms have finished.
            """
            worker_id = worker_comm.id
            task_name = self.task.unique_ref
            name = f"Task [{task_name}] (worker_id: {worker_id})"
            closed = False

            try:
                while not closed and (_msg := await host_comm.as_async.request()):
                    assert isinstance(_msg, tuple), "Expected (msg_kind, data)!"
                    msg_kind, data = _msg
                    logger.debug(f"{self.name}: receieved {msg_kind} with {data=}")

                    match msg_kind:
                        # Other side has closed the connection, break out of coroutine
                        case Comm.Msg.Kind.CLOSE:
                            closed = True
                        case Comm.Msg.Kind.CLOSE_WITH_ACK:
                            host_comm._send_pipe(Comm.Msg.Kind.ACK)
                            closed = True
                        case Comm.Msg.Kind.OPEN:
                            self.open_comms.add(worker_id)
                        case _:
                            pass

                    event = EVENT_LOOKUP[msg_kind]
                    msg = Comm.Msg(
                        comm=host_comm,
                        data=data,
                        kind=msg_kind,
                        future=future,
                        task=self.task,
                    )
                    self.task.emitter.emit(event, msg)

            except EOFError:
                # This means the connection dropped to the worker, however this is not
                # an error in the main process and so we can safely ignore that.
                logger.debug(f"{name}: closed connection")
            except Exception as e:
                # Something unexpected happened in the main process, either from us or
                # from a users callback. In this case we want to raise the exception
                logger.error(
                    f"{name}: Exception occured in scheduler or callbacks!",
                    exc_info=e,
                )

                # NOTE: It's important that we let the worker know that something went
                # wrong, especially if it's requesting things. The worker will only
                # see this msg when it does a `request()`
                host_comm._send_pipe(Comm.Msg.Kind.WORKER_CLOSE_REQUEST)
                raise e
            finally:
                # Make sure we do all the clean up!
                logger.debug(f"{name}: finished communication, closing comms")

                # We don't necessarily know how we got here but
                host_comm.close(
                    wait_for_ack=False,
                    okay_if_broken_pipe=True,
                    side="host",
                )
                worker_comm.close(
                    wait_for_ack=False,
                    okay_if_broken_pipe=True,
                    side="host-on-worker-comm",
                )

                if worker_id in self.open_comms:
                    self.open_comms.remove(worker_id)

                # Remove the reference to the work comm so it gets garbarged
                del self.comms[worker_id]
                logger.debug(f"{name}: finished and cleaned")

        @override
        def __rich__(self) -> Panel:
            from rich.panel import Panel
            from rich.text import Text

            return Panel(
                Text("Open Connections: ").append(str(len(self.open_comms)), "yellow"),
                title=f"Plugin {self.name}",
            )

    @dataclass
    class Msg(Generic[T]):
        """A message sent over a communication channel.

        Attributes:
            task: The task that sent the message.
            comm: The communication channel.
            future: The future of the task.
            data: The data sent by the task.
        """

        kind: Kind
        data: T
        comm: Comm = field(repr=False)
        future: asyncio.Future = field(repr=False)
        task: Task = field(repr=False)

        def respond(self, response: Any) -> None:
            """Respond to the message.

            Args:
                response: The response to send back to the task.
            """
            self.comm._send_pipe(response)

        class Kind(str, Enum):
            """The kind of message."""

            CLOSE = "close"
            CLOSE_WITH_ACK = "close-with-ack"
            WORKER_CLOSE_REQUEST = "worker-close-request"
            OPEN = "open"
            MESSAGE = "message"
            REQUEST = "request"
            ACK = "ack"

            @override
            def __str__(self) -> str:
                return self.value

    class TimeoutError(MPTimeoutError):  # noqa: A001
        """A timeout error for communications."""

    class CloseRequestError(RuntimeError):
        """An exception happened in the main process and it send
        a response to the worker to raise this exception.
        """


EVENT_LOOKUP = {
    Comm.Msg.Kind.CLOSE: Comm.CLOSE,
    Comm.Msg.Kind.CLOSE_WITH_ACK: Comm.CLOSE,
    Comm.Msg.Kind.OPEN: Comm.OPEN,
    Comm.Msg.Kind.MESSAGE: Comm.MESSAGE,
    Comm.Msg.Kind.REQUEST: Comm.REQUEST,
}



"""Stores low-level types used through the library."""
from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence
from dataclasses import dataclass
from itertools import chain, repeat
from typing import Any, Generic, NoReturn, Protocol, TypeAlias, TypeVar
from typing_extensions import override

import numpy as np

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")

# NOTE: We do not provide sorted types for mutable data structures
# as modifying them could potentially ruin the sorting.

SortedSequence: TypeAlias = Sequence[T]
"""A sequence that is sorted. Only useful for typing"""

SortedIterable: TypeAlias = Iterable[T]
"""An iterable that is sorted. Only useful for typing"""

Item = TypeVar("Item")
"""The type associated with components, splits and choices"""

Config: TypeAlias = Mapping[str, Any]
"""An object representing a configuration of a pipeline."""

Space = TypeVar("Space")
"""Generic for objects that are aware of a space but not the specific kind"""

Seed: TypeAlias = int | np.integer | (np.random.RandomState | np.random.Generator)
"""Type alias for kinds of Seeded objects."""

FidT: TypeAlias = tuple[int, int] | tuple[float, float] | list[Any]
"""Type alias for a fidelity bound."""


class Comparable(Protocol):
    """Protocol for annotating comparable types."""

    @abstractmethod
    def __lt__(self: _CT, other: _CT, /) -> bool:
        pass

    @abstractmethod
    def __gt__(self: _CT, other: _CT, /) -> bool:
        pass


_CT = TypeVar("_CT", bound=Comparable)


def assert_never(value: NoReturn) -> NoReturn:
    """Utility function for asserting that a value is never reached."""
    # This also works in runtime as well:
    raise AssertionError(f"This code should never be reached, got: {value}")


def safe_issubclass(cls: type, classes: str | tuple[str, ...]) -> bool:
    """Check if a class is a subclass of a given type.

    This is a safe version of issubclass that relies on strings,
    which is useful for when the type is not importable.

    Args:
        cls: The class to check
        classes: The type to check for.

    Returns:
        bool
    """

    def type_names(o: type) -> Iterator[str]:
        yield o.__qualname__
        for parent in o.__bases__:
            yield from type_names(parent)

    allowable_names = {classes} if isinstance(classes, str) else set(classes)
    return any(name in allowable_names for name in type_names(cls))


def safe_isinstance(obj: Any, t: str | tuple[str, ...]) -> bool:
    """Check if an object is of a given type.

    This is a safe version of isinstance that relies on strings,
    which is useful for when the type is not importable.

    Args:
        obj: The object to check
        t: The type to check for.

    Returns:
        bool
    """
    return safe_issubclass(type(obj), t)


class Requeue(Iterator[T]):
    """A queue that can have items requeued.

    ```python exec="true" source="material-block" result="python" title="Requeue"
    import random
    from amltk.types import Requeue

    name_generator = iter(["Alice", "Bob", "Charlie"])
    queue: Requeue[str] = Requeue(name_generator)

    rng = random.Random(1)

    def process_name(name: str) -> bool:
        return rng.choice([True, False])

    for name in queue:
        print(f"Processing {name}")
        processed = process_name(name)
        if not processed:
            print(f"Failed to process {name}, requeuing")
            queue.requeue(name)
    ```


    See Also:
        * [`Requeue.from_func(f)`][amltk.types.Requeue.from_func]

            If you have a function which will generate items, you can use
            this to create a requeue from it.

        * [`.append(item)`][amltk.types.Requeue.append]

            Append an item to the end of the queue

        * [`.requeue(item)`][amltk.types.Requeue.requeue]

            Requeue an item to the start of the queue
    """

    def __init__(self, generator: Iterable[T]) -> None:
        """Create a requeue from an iterable."""
        super().__init__()
        self.generator = iter(generator)

    @override
    def __next__(self) -> T:
        return next(self.generator)

    def append(self, item: T) -> None:
        """Append an item to the queue."""
        self.generator = chain(self.generator, [item])

    def requeue(self, item: T) -> None:
        """Requeue an item."""
        self.generator = chain([item], self.generator)

    @classmethod
    def from_func(cls, func: Callable[[], T], n: int | None = None) -> Requeue[T]:
        """Create a Requeue from a function."""
        repeater = repeat(None) if n is None else repeat(None, times=n)
        return cls(func() for _ in repeater)


@dataclass
class StoredValue(Generic[K, V]):
    """A value that is stored on disk and loaded lazily.

    This is useful for transmitting large objects between processes.

    ```python exec="true" source="material-block" result="python" title="StoredValue"
    from amltk.types import StoredValue
    import pandas as pd
    from pathlib import Path

    df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    path = Path("df.csv")
    df.to_csv(path)

    stored_value = StoredValue(path, read=pd.read_csv)

    # Somewhere in a processes
    df = stored_value.value()
    print(df)

    path.unlink()
    ```

    You can quickly obtain these from buckets if you require
    ```python exec="true" source="material-block" result="python" title="StoredValue from bucket"
    from amltk import PathBucket
    import pandas as pd

    df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    bucket = PathBucket("bucket_path")
    bucket.update({"df.csv": df})

    stored_value = bucket["df.csv"].as_stored_value()

    # Somewhere in a processes
    df = stored_value.value()
    print(df)

    bucket.rmdir()
    ```
    """  # noqa: E501

    key: K
    read: Callable[[K], V]

    _value: V | None = None

    def value(self) -> V:
        """Get the value."""
        if self._value is None:
            self._value = self.read(self.key)

        return self._value



"""Utilities for working with rich."""
# NOTE: All rich imports should be done locally to prevent any issue
# where rich not being installed.
from __future__ import annotations

from concurrent.futures import ProcessPoolExecutor
from typing import TYPE_CHECKING, Any

from amltk._richutil.renderers import (
    ProcessPoolExecutorRenderer,
    rich_make_column_selector,
)
from amltk.types import safe_isinstance

if TYPE_CHECKING:
    import pandas as pd
    from rich.console import OverflowMethod, RenderableType
    from rich.pretty import Pretty
    from rich.style import StyleType
    from rich.table import Table
    from rich.text import TextType


def richify(
    obj: Any,
    *,
    otherwise: type[Pretty] | None = None,
) -> Any | RenderableType:
    """Try to convert an object to a rich object.

    This is mainly for common objects we may encounter and
    want to prettify if we can.

    Args:
        obj: Object to convert.
        otherwise: If not None, return this called with obj.
            Else, return obj.
    """
    if safe_isinstance(obj, "make_column_selector"):
        return rich_make_column_selector(obj)

    if isinstance(obj, ProcessPoolExecutor):
        return ProcessPoolExecutorRenderer(obj)

    if otherwise:
        return otherwise(obj)

    return obj


def df_to_table(
    df: pd.DataFrame,
    *,
    title: TextType | None = None,
    index_style: StyleType = "",
    expand: bool = True,
    overflow: OverflowMethod = "ellipsis",
) -> Table:
    """Convert a dataframe to a rich table."""
    from rich.table import Column, Table

    index_name = df.index.name if df.index.name is not None else "Index"
    headers = [
        Column(str(index_name), justify="left", style=index_style, overflow=overflow),
    ]
    headers.extend(df.columns)

    table = Table(*headers, title=title, title_justify="left", expand=expand)
    for index, row in df.iterrows():
        table.add_row(str(index), *[str(cell) for cell in row])

    return table



"""This module is concerned with the termination of the workers of a scheduler.

Most Executors on `shutdown(wait=False)` will not terminate their workers
but just cancel pending futures. This means that currently running tasks
will continue to run until they finish and have their result discarded as
the program will have moved on.

We provide some custom strategies for known executors.

Note:
    There is no known way in basic Python to forcibully terminate a thread
    that does not account for early terminiation explicitly.
"""
from __future__ import annotations

from collections.abc import Callable
from concurrent.futures import Executor, ProcessPoolExecutor, ThreadPoolExecutor
from contextlib import suppress
from typing import TypeVar

import psutil

_Executor = TypeVar("_Executor", bound=Executor)


def _terminate_with_psutil(executor: ProcessPoolExecutor) -> None:
    """Terminate all processes in the given executor using psutil.

    Should work for ProcessPoolExecutor cross platform.

    Args:
        executor: The executor to terminate.
    """
    # They've already finished
    if not executor._processes:
        return

    worker_processes = [psutil.Process(p.pid) for p in executor._processes.values()]
    for worker_process in worker_processes:
        try:
            child_preocesses = worker_process.children(recursive=True)
        except psutil.NoSuchProcess:
            continue

        for child_process in child_preocesses:
            with suppress(psutil.NoSuchProcess):
                child_process.terminate()

        with suppress(psutil.NoSuchProcess):
            worker_process.terminate()


def termination_strategy(executor: _Executor) -> Callable[[_Executor], None] | None:
    """Return a termination strategy for the given executor.

    Args:
        executor: The executor to get a termination strategy for.

    Returns:
        A termination strategy for the given executor, or None if no
        termination strategy is available.
    """
    if isinstance(executor, ThreadPoolExecutor):
        return None

    if isinstance(executor, ProcessPoolExecutor):
        return _terminate_with_psutil  # type: ignore

    # Dask process based things seem pretty happy to close nicely and need
    # no special treatment.

    return None



"""A [`Trial`][amltk.optimization.Trial] is
typically the output of
[`Optimizer.ask()`][amltk.optimization.Optimizer.ask], indicating
what the optimizer would like to evaluate next. We provide a host
of convenience methods attached to the `Trial` to make it easy to
save results, store artifacts, and more.

Paired with the `Trial` is the [`Trial.Report`][amltk.optimization.Trial.Report],
class, providing an easy way to report back to the optimizer's
[`tell()`][amltk.optimization.Optimizer.tell] with
a simple [`trial.success(cost=...)`][amltk.optimization.Trial.success] or
[`trial.fail(cost=...)`][amltk.optimization.Trial.fail] call..

### Trial

::: amltk.optimization.trial.Trial
    options:
        members: False

### Report

::: amltk.optimization.trial.Trial.Report
    options:
        members: False

"""
from __future__ import annotations

import copy
import logging
import traceback
from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generic, Literal, TypeVar, overload
from typing_extensions import ParamSpec, Self, override

import numpy as np
import pandas as pd

from amltk._functional import dict_get_not_none, mapping_select, prefix_keys
from amltk._richutil.renderable import RichRenderable
from amltk.optimization.metric import Metric
from amltk.profiling import Memory, Profile, Profiler, Timer
from amltk.store import Bucket, PathBucket

if TYPE_CHECKING:
    from rich.console import RenderableType
    from rich.panel import Panel
    from rich.text import Text

# Inner trial info object
I = TypeVar("I")  # noqa: E741
I2 = TypeVar("I2")

# Parameters to task
P = ParamSpec("P")
P2 = ParamSpec("P2")

# Return type of task
R = TypeVar("R")

# Random TypeVar
T = TypeVar("T")

logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class Trial(RichRenderable, Generic[I]):
    """A [`Trial`][amltk.optimization.Trial] encapsulates some configuration
    that needs to be evaluated. Typically this is what is generated by an
    [`Optimizer.ask()`][amltk.optimization.Optimizer.ask] call.

    ??? tip "Usage"

        To begin a trial, you can use the
        [`trial.begin()`][amltk.optimization.Trial.begin], which will catch
        exceptions/traceback and profile the block of code.

        If all went smooth, your trial was successful and you can use
        [`trial.success()`][amltk.optimization.Trial.success] to generate
        a success [`Report`][amltk.optimization.Trial.Report], typically
        passing what your chosen optimizer expects, e.g. `"loss"` or `"cost"`.

        If your trial failed, you can instead use the
        [`trial.fail()`][amltk.optimization.Trial.fail] to generate a
        failure [`Report`][amltk.optimization.Trial.Report], where
        any caught exception will be attached to it. Each
        [`Optimizer`][amltk.optimization.Optimizer] will take care of what to do
        from here.

        ```python exec="true" source="material-block" html="true"
        from amltk.optimization import Trial, Metric
        from amltk.store import PathBucket

        cost = Metric("cost", minimize=True)

        def target_function(trial: Trial) -> Trial.Report:
            x = trial.config["x"]
            y = trial.config["y"]

            with trial.begin():
                cost = x**2 - y

            if trial.exception:
                return trial.fail()

            return trial.success(cost=cost)

        # ... usually obtained from an optimizer
        trial = Trial(name="some-unique-name", config={"x": 1, "y": 2}, metrics=[cost])

        report = target_function(trial)
        print(report.df())
        ```


    What you can return with [`trial.success()`][amltk.optimization.Trial.success]
    or [`trial.fail()`][amltk.optimization.Trial.fail] depends on the
    [`metrics`][amltk.optimization.Trial.metrics] of the trial. Typically
    an optimizer will provide the trial with the list of metrics.

    ??? tip "Metrics"

        ::: amltk.optimization.metric.Metric
            options:
                members: False

    Some important properties is that they have a unique
    [`.name`][amltk.optimization.Trial.name] given the optimization run,
    a candidate [`.config`][amltk.optimization.Trial.config]' to evaluate,
    a possible [`.seed`][amltk.optimization.Trial.seed] to use,
    and an [`.info`][amltk.optimization.Trial.info] object which is the optimizer
    specific information, if required by you.

    If using [`Plugins`][amltk.scheduling.plugins.Plugin], they may insert
    some extra objects in the [`.extra`][amltk.optimization.Trial.extras] dict.

    To profile your trial, you can wrap the logic you'd like to check with
    [`trial.begin()`][amltk.optimization.Trial.begin], which will automatically
    catch any errors, record the traceback, and profile the block of code, in
    terms of time and memory.

    You can access the profiled time and memory using the
    [`.time`][amltk.optimization.Trial.time] and
    [`.memory`][amltk.optimization.Trial.memory] attributes.
    If you've [`profile()`][amltk.optimization.Trial.profile]'ed any other intervals,
    you can access them by name through
    [`trial.profiles`][amltk.optimization.Trial.profiles].
    Please see the [`Profiler`][amltk.profiling.profiler.Profiler]
    for more.

    ??? example "Profiling with a trial."

        ```python exec="true" source="material-block" result="python" title="profile"
        from amltk.optimization import Trial

        trial = Trial(name="some-unique-name", config={})

        # ... somewhere where you've begun your trial.
        with trial.profile("some_interval"):
            for work in range(100):
                pass

        print(trial.profiler.df())
        ```

    You can also record anything you'd like into the
    [`.summary`][amltk.optimization.Trial.summary], a plain `#!python dict`
    or use [`trial.store()`][amltk.optimization.Trial.store] to store artifacts
    related to the trial.

    ??? tip "What to put in `.summary`?"

        For large items, e.g. predictions or models, these are highly advised to
        [`.store()`][amltk.optimization.Trial.store] to disk, especially if using
        a `Task` for multiprocessing.

        Further, if serializing the report using the
        [`report.df()`][amltk.optimization.Trial.Report.df],
        returning a single row,
        or a [`History`][amltk.optimization.History]
        with [`history.df()`][amltk.optimization.History.df] for a dataframe consisting
        of many of the reports, then you'd likely only want to store things
        that are scalar and can be serialised to disk by a pandas DataFrame.

    """

    name: str
    """The unique name of the trial."""

    config: Mapping[str, Any]
    """The config of the trial provided by the optimizer."""

    bucket: PathBucket = field(
        default_factory=lambda: PathBucket("unknown-trial-bucket"),
    )
    """The bucket to store trial related output to."""

    info: I | None = field(default=None, repr=False)
    """The info of the trial provided by the optimizer."""

    metrics: Sequence[Metric] = field(default_factory=list)
    """The metrics associated with the trial."""

    seed: int | None = None
    """The seed to use if suggested by the optimizer."""

    fidelities: dict[str, Any] | None = None
    """The fidelities at which to evaluate the trial, if any."""

    time: Timer.Interval = field(repr=False, default_factory=Timer.na)
    """The time taken by the trial, once ended."""

    memory: Memory.Interval = field(repr=False, default_factory=Memory.na)
    """The memory used by the trial, once ended."""

    profiler: Profiler = field(
        repr=False,
        default_factory=lambda: Profiler(memory_unit="B", time_kind="wall"),
    )
    """A profiler for this trial."""

    summary: dict[str, Any] = field(default_factory=dict)
    """The summary of the trial. These are for summary statistics of a trial and
    are single values."""

    exception: BaseException | None = field(repr=True, default=None)
    """The exception raised by the trial, if any."""

    traceback: str | None = field(repr=False, default=None)
    """The traceback of the exception, if any."""

    storage: set[Any] = field(default_factory=set)
    """Anything stored in the trial, the elements of the list are keys that can be
    used to retrieve them later, such as a Path.
    """

    extras: dict[str, Any] = field(default_factory=dict)
    """Any extras attached to the trial."""

    @property
    def profiles(self) -> Mapping[str, Profile.Interval]:
        """The profiles of the trial."""
        return self.profiler.profiles

    @contextmanager
    def begin(
        self,
        time: Timer.Kind | Literal["wall", "cpu", "process"] | None = None,
        memory_unit: Memory.Unit | Literal["B", "KB", "MB", "GB"] | None = None,
    ) -> Iterator[None]:
        """Begin the trial with a `contextmanager`.

        Will begin timing the trial in the `with` block, attaching the profiled time and memory
        to the trial once completed, under `.profile.time` and `.profile.memory` attributes.

        If an exception is raised, it will be attached to the trial under `.exception`
        with the traceback attached to the actual error message, such that it can
        be pickled and sent back to the main process loop.

        ```python exec="true" source="material-block" result="python" title="begin" hl_lines="5"
        from amltk.optimization import Trial

        trial = Trial(name="trial", config={"x": 1})

        with trial.begin():
            # Do some work
            pass

        print(trial.memory)
        print(trial.time)
        ```

        ```python exec="true" source="material-block" result="python" title="begin-fail" hl_lines="5"
        from amltk.optimization import Trial

        trial = Trial(name="trial", config={"x": -1})

        with trial.begin():
            raise ValueError("x must be positive")

        print(trial.exception)
        print(trial.traceback)
        print(trial.memory)
        print(trial.time)
        ```

        Args:
            time: The timer kind to use for the trial. Defaults to the default
                timer kind of the profiler.
            memory_unit: The memory unit to use for the trial. Defaults to the
                default memory unit of the profiler.
        """  # noqa: E501
        with self.profiler(name="trial", memory_unit=memory_unit, time_kind=time):
            try:
                yield
            except Exception as error:  # noqa: BLE001
                self.exception = error
                self.traceback = traceback.format_exc()
            finally:
                self.time = self.profiler["trial"].time
                self.memory = self.profiler["trial"].memory

    @contextmanager
    def profile(
        self,
        name: str,
        *,
        time: Timer.Kind | Literal["wall", "cpu", "process"] | None = None,
        memory_unit: Memory.Unit | Literal["B", "KB", "MB", "GB"] | None = None,
        summary: bool = False,
    ) -> Iterator[None]:
        """Measure some interval in the trial.

        The results of the profiling will be available in the `.summary` attribute
        with the name of the interval as the key.

        ```python exec="true" source="material-block" result="python" title="profile"
        from amltk.optimization import Trial
        import time

        trial = Trial(name="trial", config={"x": 1})

        with trial.profile("some_interval"):
            # Do some work
            time.sleep(1)

        print(trial.profiler["some_interval"].time)
        ```

        Args:
            name: The name of the interval.
            time: The timer kind to use for the trial. Defaults to the default
                timer kind of the profiler.
            memory_unit: The memory unit to use for the trial. Defaults to the
                default memory unit of the profiler.
            summary: Whether to add the interval to the summary.

        Yields:
            The interval measured. Values will be nan until the with block is finished.
        """
        with self.profiler(name=name, memory_unit=memory_unit, time_kind=time):
            yield

        if summary:
            profile = self.profiler[name]
            self.summary.update(profile.to_dict(prefix=name))

    def success(self, **metrics: float | int) -> Trial.Report[I]:
        """Generate a success report.

        ```python exec="true" source="material-block" result="python" title="success" hl_lines="7"
        from amltk.optimization import Trial, Metric

        loss_metric = Metric("loss", minimize=True)

        trial = Trial(name="trial", config={"x": 1}, metrics=[loss_metric])

        with trial.begin():
            # Do some work
            report = trial.success(loss=1)

        print(report)
        ```

        Args:
            **metrics: The metrics of the trial, where the key is the name of the
                metrics and the value is the metric.

        Returns:
            The report of the trial.
        """  # noqa: E501
        _recorded_values: list[Metric.Value] = []
        for _metric in self.metrics:
            if (raw_value := metrics.get(_metric.name)) is not None:
                _recorded_values.append(_metric.as_value(raw_value))
            else:
                raise ValueError(
                    f"Cannot report success without {self.metrics=}."
                    f" Please provide a value for the metric.",
                )

        # Need to check if anything extra was reported!
        extra = set(metrics.keys()) - {metric.name for metric in self.metrics}
        if extra:
            raise ValueError(
                f"Cannot report success with extra metrics: {extra=}."
                f"\nOnly {self.metrics=} are allowed.",
            )

        return Trial.Report(
            trial=self,
            status=Trial.Status.SUCCESS,
            metric_values=tuple(_recorded_values),
        )

    def fail(self, **metrics: float | int) -> Trial.Report[I]:
        """Generate a failure report.

        !!! note "Non specifed metrics"

            If you do not specify metrics, this will use
            the [`.metrics`][amltk.optimization.Trial.metrics] to determine
            the [`.worst`][amltk.optimization.Metric.worst] value of the metric,
            using that as the reported result

        ```python exec="true" source="material-block" result="python" title="fail"
        from amltk.optimization import Trial, Metric

        loss = Metric("loss", minimize=True, bounds=(0, 1_000))
        trial = Trial(name="trial", config={"x": 1}, metrics=[loss])

        with trial.begin():
            raise ValueError("This is an error")  # Something went wrong

        if trial.exception: # You can check for an exception of the trial here
            report = trial.fail()

        print(report.metrics)
        print(report)
        ```

        Returns:
            The result of the trial.
        """
        _recorded_values: list[Metric.Value] = []
        for _metric in self.metrics:
            if (raw_value := metrics.get(_metric.name)) is not None:
                _recorded_values.append(_metric.as_value(raw_value))
            else:
                _recorded_values.append(_metric.worst)

        return Trial.Report(
            trial=self,
            status=Trial.Status.FAIL,
            metric_values=tuple(_recorded_values),
        )

    def crashed(
        self,
        exception: BaseException | None = None,
        traceback: str | None = None,
    ) -> Trial.Report[I]:
        """Generate a crash report.

        !!! note

            You will typically not create these manually, but instead if we don't
            recieve a report from a target function evaluation, but only an error,
            we assume something crashed and generate a crash report for you.

        !!! note "Non specifed metrics"

            We will use the [`.metrics`][amltk.optimization.Trial.metrics] to determine
            the [`.worst`][amltk.optimization.Metric.worst] value of the metric,
            using that as the reported metrics

        Args:
            exception: The exception that caused the crash. If not provided, the
                exception will be taken from the trial. If this is still `None`,
                a `RuntimeError` will be raised.
            traceback: The traceback of the exception. If not provided, the
                traceback will be taken from the trial if there is one there.

        Returns:
            The report of the trial.
        """
        if exception is None and self.exception is None:
            raise RuntimeError(
                "Cannot generate a crash report without an exception."
                " Please provide an exception or use `with trial.begin():` to start"
                " the trial.",
            )

        self.exception = exception if exception else self.exception
        self.traceback = traceback if traceback else self.traceback

        return Trial.Report(
            trial=self,
            status=Trial.Status.CRASHED,
            metric_values=tuple(metric.worst for metric in self.metrics),
        )

    def store(
        self,
        items: Mapping[str, T],
        *,
        where: (
            str | Path | Bucket | Callable[[str, Mapping[str, T]], None] | None
        ) = None,
    ) -> None:
        """Store items related to the trial.

        ```python exec="true" source="material-block" result="python" title="store" hl_lines="5"
        from amltk.optimization import Trial
        from amltk.store import PathBucket

        trial = Trial(name="trial", config={"x": 1}, bucket=PathBucket("results"))
        trial.store({"config.json": trial.config})

        print(trial.storage)
        ```

        You could also specify `where=` exactly to store the thing

        ```python exec="true" source="material-block" result="python" title="store-bucket" hl_lines="7"
        from amltk.optimization import Trial

        trial = Trial(name="trial", config={"x": 1})
        trial.store({"config.json": trial.config}, where="./results")

        print(trial.storage)
        ```

        Args:
            items: The items to store, a dict from the key to store it under
                to the item itself.If using a `str`, `Path` or `PathBucket`,
                the keys of the items should be a valid filename, including
                the correct extension. e.g. `#!python {"config.json": trial.config}`

            where: Where to store the items.

                * If `None`, will use the bucket attached to the `Trial` if any,
                    otherwise it will raise an error.

                * If a `str` or `Path`, will store
                a bucket will be created at the path, and the items will be
                stored in a sub-bucket with the name of the trial.

                * If a `Bucket`, will store the items **in a sub-bucket** with the
                name of the trial.

                * If a `Callable`, will call the callable with the name of the
                trial and the key-valued pair of items to store.
        """  # noqa: E501
        method: Bucket
        match where:
            case None:
                method = self.bucket
                method.sub(self.name).store(items)
            case str() | Path():
                method = PathBucket(where, create=True)
                method.sub(self.name).store(items)
            case Bucket():
                method = where
                method.sub(self.name).store(items)
            case _:
                # Leave it up to supplied method
                where(self.name, items)

        # Add the keys to storage
        self.storage.update(items.keys())

    def delete_from_storage(
        self,
        items: Iterable[str],
        *,
        where: (
            str | Path | Bucket | Callable[[str, Iterable[str]], dict[str, bool]] | None
        ) = None,
    ) -> dict[str, bool]:
        """Delete items related to the trial.

        ```python exec="true" source="material-block" result="python" title="delete-storage" hl_lines="6"
        from amltk.optimization import Trial
        from amltk.store import PathBucket

        bucket = PathBucket("results")
        trial = Trial(name="trial", config={"x": 1}, info={}, bucket=bucket)

        trial.store({"config.json": trial.config})
        trial.delete_from_storage(items=["config.json"])

        print(trial.storage)
        ```

        You could also create a Bucket and use that instead.

        ```python exec="true" source="material-block" result="python" title="delete-storage-bucket" hl_lines="9"
        from amltk.optimization import Trial
        from amltk.store import PathBucket

        bucket = PathBucket("results")
        trial = Trial(name="trial", config={"x": 1}, bucket=bucket)

        trial.store({"config.json": trial.config})
        trial.delete_from_storage(items=["config.json"])

        print(trial.storage)
        ```

        Args:
            items: The items to delete, an iterable of keys
            where: Where the items are stored

                * If `None`, will use the bucket attached to the `Trial` if any,
                    otherwise it will raise an error.

                * If a `str` or `Path`, will lookup a bucket at the path,
                and the items will be deleted from a sub-bucket with the name of the trial.

                * If a `Bucket`, will delete the items in a sub-bucket with the
                name of the trial.

                * If a `Callable`, will call the callable with the name of the
                trial and the keys of the items to delete. Should a mapping from
                the key to whether it was deleted or not.

        Returns:
            A dict from the key to whether it was deleted or not.
        """  # noqa: E501
        # If not a Callable, we convert to a path bucket
        method: Bucket
        match where:
            case None:
                method = self.bucket
            case str() | Path():
                method = PathBucket(where, create=False)
            case Bucket():
                method = where
            case _:
                # Leave it up to supplied method
                return where(self.name, items)

        sub_bucket = method.sub(self.name)
        return sub_bucket.remove(items)

    def copy(self) -> Self:
        """Create a copy of the trial.

        Returns:
            The copy of the trial.
        """
        return copy.deepcopy(self)

    @overload
    def retrieve(
        self,
        key: str,
        *,
        where: str | Path | Bucket[str, Any] | None = ...,
        check: None = None,
    ) -> Any:
        ...

    @overload
    def retrieve(
        self,
        key: str,
        *,
        where: str | Path | Bucket[str, Any] | None = ...,
        check: type[R],
    ) -> R:
        ...

    def retrieve(
        self,
        key: str,
        *,
        where: str | Path | Bucket[str, Any] | None = None,
        check: type[R] | None = None,
    ) -> R | Any:
        """Retrieve items related to the trial.

        !!! note "Same argument for `where=`"

             Use the same argument for `where=` as you did for `store()`.

        ```python exec="true" source="material-block" result="python" title="retrieve" hl_lines="7"
        from amltk.optimization import Trial
        from amltk.store import PathBucket

        bucket = PathBucket("results")

        # Create a trial, normally done by an optimizer
        trial = Trial(name="trial", config={"x": 1}, bucket=bucket)

        trial.store({"config.json": trial.config})
        config = trial.retrieve("config.json")

        print(config)
        ```

        You could also manually specify where something get's stored and retrieved

        ```python exec="true" source="material-block" result="python" title="retrieve-bucket" hl_lines="11"

        from amltk.optimization import Trial
        from amltk.store import PathBucket

        path = "./config_path"

        trial = Trial(name="trial", config={"x": 1})

        trial.store({"config.json": trial.config}, where=path)

        config = trial.retrieve("config.json", where=path)
        print(config)
        import shutil; shutil.rmtree(path)  # markdown-exec: hide
        ```

        Args:
            key: The key of the item to retrieve as said in `.storage`.
            check: If provided, will check that the retrieved item is of the
                provided type. If not, will raise a `TypeError`. This
                is only used if `where=` is a `str`, `Path` or `Bucket`.

            where: Where to retrieve the items from.

                * If `None`, will use the bucket attached to the `Trial` if any,
                    otherwise it will raise an error.

                * If a `str` or `Path`, will store
                a bucket will be created at the path, and the items will be
                retrieved from a sub-bucket with the name of the trial.

                * If a `Bucket`, will retrieve the items from a sub-bucket with the
                name of the trial.

        Returns:
            The retrieved item.

        Raises:
            TypeError: If `check=` is provided and  the retrieved item is not of the provided
                type.
        """  # noqa: E501
        # If not a Callable, we convert to a path bucket
        method: Bucket[str, Any]
        match where:
            case None:
                method = self.bucket
            case str():
                method = PathBucket(where, create=True)
            case Path():
                method = PathBucket(where, create=True)
            case Bucket():
                method = where

        # Store in a sub-bucket
        return method.sub(self.name)[key].load(check=check)

    def attach_extra(self, name: str, plugin_item: Any) -> None:
        """Attach a plugin item to the trial.

        Args:
            name: The name of the plugin item.
            plugin_item: The plugin item.
        """
        self.extras[name] = plugin_item

    def rich_renderables(self) -> Iterable[RenderableType]:  # noqa: C901
        """The renderables for rich for this report."""
        from rich.panel import Panel
        from rich.pretty import Pretty
        from rich.table import Table
        from rich.text import Text

        items = []
        table = Table.grid(padding=(0, 1), expand=False)

        # Predfined things
        table.add_row("config", Pretty(self.config))

        if self.fidelities:
            table.add_row("fidelities", Pretty(self.fidelities))

        if any(self.extras):
            table.add_row("extras", Pretty(self.extras))

        if self.seed:
            table.add_row("seed", Pretty(self.seed))

        if self.bucket:
            table.add_row("bucket", Pretty(self.bucket))

        if self.metrics:
            items.append(
                Panel(Pretty(self.metrics), title="Metrics", title_align="left"),
            )

        # Dynamic things
        if self.summary:
            table.add_row("summary", Pretty(self.summary))

        if any(self.storage):
            table.add_row("storage", Pretty(self.storage))

        if self.exception:
            table.add_row("exception", Text(str(self.exception), style="bold red"))

        if self.traceback:
            table.add_row("traceback", Text(self.traceback, style="bold red"))

        for name, profile in self.profiles.items():
            table.add_row("profile:" + name, Pretty(profile))

        yield from items

    @override
    def __rich__(self) -> RenderableType:
        from rich.console import Group as RichGroup
        from rich.panel import Panel
        from rich.text import Text

        title = Text.assemble(
            ("Trial", "bold"),
            ("(", "default"),
            (self.name, "italic"),
            (")", "default"),
        )

        return Panel(
            RichGroup(*self.rich_renderables()),
            title=title,
            title_align="left",
        )

    class Status(str, Enum):
        """The status of a trial."""

        SUCCESS = "success"
        """The trial was successful."""

        FAIL = "fail"
        """The trial failed."""

        CRASHED = "crashed"
        """The trial crashed."""

        UNKNOWN = "unknown"
        """The status of the trial is unknown."""

        @override
        def __str__(self) -> str:
            return self.value

        def __rich__(self) -> Text:
            from rich.text import Text

            styles = {
                Trial.Status.SUCCESS: "bold green",
                Trial.Status.FAIL: "bold yellow",
                Trial.Status.CRASHED: "bold red",
                Trial.Status.UNKNOWN: "bold underline",
            }

            return Text(self.value, style=styles.get(self, "bold underline"))

    @dataclass
    class Report(RichRenderable, Generic[I2]):
        """The [`Trial.Report`][amltk.optimization.Trial.Report] encapsulates
        a [`Trial`][amltk.optimization.Trial], its status and any metrics/exceptions
        that may have occured.

        Typically you will not create these yourself, but instead use
        [`trial.success()`][amltk.optimization.Trial.success] or
        [`trial.fail()`][amltk.optimization.Trial.fail] to generate them.

        ```python exec="true" source="material-block" result="python"
        from amltk.optimization import Trial, Metric

        loss = Metric("loss", minimize=True)

        trial = Trial(name="trial", config={"x": 1}, metrics=[loss])

        with trial.begin():
            # Do some work
            # ...
            report: Trial.Report = trial.success(loss=1)

        print(report.df())
        ```

        These reports are used to report back metrics to an
        [`Optimizer`][amltk.optimization.Optimizer]
        with [`Optimizer.tell()`][amltk.optimization.Optimizer.tell] but can also be
        stored for your own uses.

        You can access the original trial with the
        [`.trial`][amltk.optimization.Trial.Report.trial] attribute, and the
        [`Status`][amltk.optimization.Trial.Status] of the trial with the
        [`.status`][amltk.optimization.Trial.Report.status] attribute.

        You may also want to check out the [`History`][amltk.optimization.History] class
        for storing a collection of `Report`s, allowing for an easier time to convert
        them to a dataframe or perform some common Hyperparameter optimization parsing
        of metrics.
        """

        trial: Trial[I2]
        """The trial that was run."""

        status: Trial.Status
        """The status of the trial."""

        metrics: dict[str, float] = field(init=False)
        """The metric values of the trial."""

        metric_values: tuple[Metric.Value, ...] = field(default_factory=tuple)
        """The metrics of the trial, linked to the metrics."""

        metric_defs: dict[str, Metric] = field(init=False)
        """A lookup to the metric definitions"""

        metric_names: tuple[str, ...] = field(init=False)
        """The names of the metrics."""

        def __post_init__(self) -> None:
            self.metrics = {value.name: value.value for value in self.metric_values}
            self.metric_names = tuple(metric.name for metric in self.metric_values)
            self.metric_defs = {v.metric.name: v.metric for v in self.metric_values}

        @property
        def exception(self) -> BaseException | None:
            """The exception of the trial, if any."""
            return self.trial.exception

        @property
        def traceback(self) -> str | None:
            """The traceback of the trial, if any."""
            return self.trial.traceback

        @property
        def name(self) -> str:
            """The name of the trial."""
            return self.trial.name

        @property
        def config(self) -> Mapping[str, Any]:
            """The config of the trial."""
            return self.trial.config

        @property
        def profiles(self) -> Mapping[str, Profile.Interval]:
            """The profiles of the trial."""
            return self.trial.profiles

        @property
        def summary(self) -> dict[str, Any]:
            """The summary of the trial."""
            return self.trial.summary

        @property
        def storage(self) -> set[str]:
            """The storage of the trial."""
            return self.trial.storage

        @property
        def time(self) -> Timer.Interval:
            """The time of the trial."""
            return self.trial.time

        @property
        def memory(self) -> Memory.Interval:
            """The memory of the trial."""
            return self.trial.memory

        @property
        def bucket(self) -> PathBucket:
            """The bucket attached to the trial."""
            return self.trial.bucket

        @property
        def info(self) -> I2 | None:
            """The info of the trial, specific to the optimizer that issued it."""
            return self.trial.info

        def df(
            self,
            *,
            profiles: bool = True,
            configs: bool = True,
            summary: bool = True,
            metrics: bool = True,
        ) -> pd.DataFrame:
            """Get a dataframe of the trial.

            !!! note "Prefixes"

                * `summary`: Entries will be prefixed with `#!python "summary:"`
                * `config`: Entries will be prefixed with `#!python "config:"`
                * `storage`: Entries will be prefixed with `#!python "storage:"`
                * `metrics`: Entries will be prefixed with `#!python "metrics:"`
                * `profile:<name>`: Entries will be prefixed with
                    `#!python "profile:<name>:"`

            Args:
                profiles: Whether to include the profiles.
                configs: Whether to include the configs.
                summary: Whether to include the summary.
                metrics: Whether to include the metrics.
            """
            items = {
                "name": self.name,
                "status": str(self.status),
                "trial_seed": self.trial.seed if self.trial.seed else np.nan,
                "exception": str(self.exception) if self.exception else "NA",
                "traceback": str(self.traceback) if self.traceback else "NA",
                "bucket": str(self.bucket.path),
            }
            if metrics:
                for value in self.metric_values:
                    items[f"metric:{value.metric}"] = value.value
            if summary:
                items.update(**prefix_keys(self.trial.summary, "summary:"))
            if configs:
                items.update(**prefix_keys(self.trial.config, "config:"))
            if profiles:
                for name, profile in sorted(self.profiles.items(), key=lambda x: x[0]):
                    # We log this one seperatly
                    if name == "trial":
                        items.update(profile.to_dict())
                    else:
                        items.update(profile.to_dict(prefix=f"profile:{name}"))

            return pd.DataFrame(items, index=[0]).convert_dtypes().set_index("name")

        @overload
        def retrieve(
            self,
            key: str,
            *,
            where: str | Path | Bucket[str, Any] | None = ...,
            check: None = None,
        ) -> Any:
            ...

        @overload
        def retrieve(
            self,
            key: str,
            *,
            where: str | Path | Bucket[str, Any] | None = ...,
            check: type[R],
        ) -> R:
            ...

        def retrieve(
            self,
            key: str,
            *,
            where: str | Path | Bucket[str, Any] | None = None,
            check: type[R] | None = None,
        ) -> R | Any:
            """Retrieve items related to the trial.

            !!! note "Same argument for `where=`"

                 Use the same argument for `where=` as you did for `store()`.

            ```python exec="true" source="material-block" result="python" title="retrieve" hl_lines="7"
            from amltk.optimization import Trial
            from amltk.store import PathBucket

            bucket = PathBucket("results")
            trial = Trial(name="trial", config={"x": 1}, bucket=bucket)

            trial.store({"config.json": trial.config})
            with trial.begin():
                report = trial.success()

            config = report.retrieve("config.json")
            print(config)
            ```

            You could also create a Bucket and use that instead.

            ```python exec="true" source="material-block" result="python" title="retrieve-bucket" hl_lines="11"

            from amltk.optimization import Trial
            from amltk.store import PathBucket

            bucket = PathBucket("results")

            trial = Trial(name="trial", config={"x": 1}, bucket=bucket)

            trial.store({"config.json": trial.config})

            with trial.begin():
                report = trial.success()

            config = report.retrieve("config.json")
            print(config)
            ```

            Args:
                key: The key of the item to retrieve as said in `.storage`.
                check: If provided, will check that the retrieved item is of the
                    provided type. If not, will raise a `TypeError`. This
                    is only used if `where=` is a `str`, `Path` or `Bucket`.
                where: Where to retrieve the items from.

                    * If `None`, will use the bucket attached to the `Trial` if any,
                        otherwise it will raise an error.

                    * If a `str` or `Path`, will store
                    a bucket will be created at the path, and the items will be
                    retrieved from a sub-bucket with the name of the trial.

                    * If a `Bucket`, will retrieve the items from a sub-bucket with the
                    name of the trial.

            Returns:
                The retrieved item.

            Raises:
                TypeError: If `check=` is provided and  the retrieved item is not of the provided
                    type.
            """  # noqa: E501
            return self.trial.retrieve(key, where=where, check=check)

        def store(
            self,
            items: Mapping[str, T],
            *,
            where: (
                str | Path | Bucket | Callable[[str, Mapping[str, T]], None] | None
            ) = None,
        ) -> None:
            """Store items related to the trial.

            See: [`Trial.store()`][amltk.optimization.trial.Trial.store]
            """
            self.trial.store(items, where=where)

        @classmethod
        def from_df(cls, df: pd.DataFrame | pd.Series) -> Trial.Report:
            """Create a report from a dataframe.

            See Also:
                * [`.from_dict()`][amltk.optimization.Trial.Report.from_dict]
            """
            if isinstance(df, pd.DataFrame):
                if len(df) != 1:
                    raise ValueError(
                        f"Expected a dataframe with one row, got {len(df)} rows.",
                    )
                series = df.iloc[0]
            else:
                series = df

            data_dict = {"name": series.name, **series.to_dict()}
            return cls.from_dict(data_dict)

        @classmethod
        def from_dict(cls, d: Mapping[str, Any]) -> Trial.Report:
            """Create a report from a dictionary.

            !!! note "Prefixes"

                Please see [`.df()`][amltk.optimization.Trial.Report.df]
                for information on what the prefixes should be for certain fields.

            Args:
                d: The dictionary to create the report from.

            Returns:
                The created report.
            """
            prof_dict = mapping_select(d, "profile:")
            if any(prof_dict):
                profile_names = sorted(
                    {name.rsplit(":", maxsplit=2)[0] for name in prof_dict},
                )
                profiles = {
                    name: Profile.from_dict(mapping_select(prof_dict, f"{name}:"))
                    for name in profile_names
                }
            else:
                profiles = {}

            # NOTE: We assume the order of the objectives are in the right
            # order in the dict. If we attempt to force a sort-order, we may
            # deserialize incorrectly. By not having a sort order, we rely
            # on serialization to keep the order, which is not ideal either.
            # May revisit this if we need to
            raw_metrics: dict[str, float] = mapping_select(d, "metric:")
            _intermediate = {
                Metric.from_str(name): value for name, value in raw_metrics.items()
            }
            metrics: dict[Metric, Metric.Value] = {
                metric: metric.as_value(value)
                for metric, value in _intermediate.items()
            }

            _trial_profile_items = {
                k: v for k, v in d.items() if k.startswith(("memory:", "time:"))
            }
            if any(_trial_profile_items):
                trial_profile = Profile.from_dict(_trial_profile_items)
                profiles["trial"] = trial_profile
            else:
                trial_profile = Profile.na()

            exception = d.get("exception")
            traceback = d.get("traceback")
            trial_seed = d.get("trial_seed")
            if pd.isna(exception) or exception == "NA":  # type: ignore
                exception = None
            if pd.isna(traceback) or traceback == "NA":  # type: ignore
                traceback = None
            if pd.isna(trial_seed):  # type: ignore
                trial_seed = None

            if (_bucket := d.get("bucket")) is not None:
                bucket = PathBucket(_bucket)
            else:
                bucket = PathBucket(f"uknown_trial_bucket-{datetime.now().isoformat()}")

            trial: Trial[None] = Trial(
                name=d["name"],
                config=mapping_select(d, "config:"),
                info=None,  # We don't save this to disk so we load it back as None
                bucket=bucket,
                seed=trial_seed,
                fidelities=mapping_select(d, "fidelities:"),
                time=trial_profile.time,
                memory=trial_profile.memory,
                profiler=Profiler(profiles=profiles),
                metrics=list(metrics.keys()),
                summary=mapping_select(d, "summary:"),
                exception=exception,
                traceback=traceback,
            )
            status = Trial.Status(dict_get_not_none(d, "status", "unknown"))
            _values: dict[str, float] = {m.name: r.value for m, r in metrics.items()}
            if status == Trial.Status.SUCCESS:
                return trial.success(**_values)

            if status == Trial.Status.FAIL:
                return trial.fail(**_values)

            if status == Trial.Status.CRASHED:
                return trial.crashed(
                    exception=Exception("Unknown status.")
                    if trial.exception is None
                    else None,
                )

            return trial.crashed(exception=Exception("Unknown status."))

        def rich_renderables(self) -> Iterable[RenderableType]:
            """The renderables for rich for this report."""
            from rich.pretty import Pretty
            from rich.text import Text

            yield Text.assemble(
                ("Status", "bold"),
                ("(", "default"),
                self.status.__rich__(),
                (")", "default"),
            )
            yield Pretty(self.metrics)
            yield from self.trial.rich_renderables()

        @override
        def __rich__(self) -> Panel:
            from rich.console import Group as RichGroup
            from rich.panel import Panel
            from rich.text import Text

            title = Text.assemble(
                ("Report", "bold"),
                ("(", "default"),
                (self.name, "italic"),
                (") - ", "default"),
                self.status.__rich__(),
            )
            return Panel(RichGroup(*self.rich_renderables()), title=title)
